
import os
import sys
import json
import ast
import re
import configparser
from io import StringIO


def parse_md(body):
    # remove comments
    pattern = r'<!---(.*?)--->'

    # Remove comments using re.sub
    body = re.sub(r'/r/n', r'/n', re.sub(pattern, '', body, flags=re.DOTALL))

    config_str = re.search(r'```\sconfigfile(.*?)```',
                           body, re.DOTALL).group(1)
    print(config_str)

    # Create a file-like object from the string
    config_file = StringIO(config_str)

    # Create a ConfigParser object
    config = configparser.ConfigParser()

    # Read configuration from the file-like object
    config.read_file(config_file)

    # Initialize an empty dictionary to hold the configuration data
    config_dict = {}

    # Iterate over sections and options
    for section in config.sections():
        config_dict[section] = {}
        for option in config.options(section):
            config_dict[section][option] = ast.literal_eval(
                config.get(section, option))

    return config_dict


def dispatch(token, payload, repo):

    import json
    from urllib import request

    # Construct the request headers
    headers = {
        "Accept": "application/vnd.github.everest-preview+json",
        "Authorization": f"token {token}",
        "Content-Type": "application/json"
    }

    # Encode the payload
    datapayload = json.dumps(payload).encode('utf-8')

    # Make the POST request
    req = request.Request(f"{repo}/dispatches",
                          data=datapayload, headers=headers, method='POST')

    # Perform the request
    try:
        with request.urlopen(req) as response:
            if response.getcode() == 204:
                print("Dispatch event triggered successfully.")
            else:
                print(
                    f"Failed to trigger dispatch event. Status code: {response.getcode()}")
                print(response.read().decode('utf-8'))
    except Exception as e:
        print(f"Error: {e}")



import os


def newrelease(owner, repo, version, content, title=''):
    print('release')
    print(os.popen(f'gh repo set-default {owner}/{repo}').read())
    release = f'gh release create "{version}" -n "{content}" -t "{title}"'
    print(os.popen(release).read())


# import json
# import urllib.request

# def trigger_CV(owner,repo,token):

#     repository = "owner/repo"
#     workflow_name = "Generate CV file"
#     branch = "main"


#      # Encode parts of the URL
#     encoded_repository = urllib.parse.quote(repository, safe='')
#     encoded_workflow_name = urllib.parse.quote(workflow_name, safe='')

#     url = f"https://api.github.com/repos/{encoded_repository}/actions/workflows/{encoded_workflow_name}/dispatches"
#     headers = {
#         "Authorization": f"Bearer {token}",
#         "Accept": "application/vnd.github.v3+json",
#         "Content-Type": "application/json"
#     }
#     payload = {
#         "ref": branch
#     }

#     # Encode payload as JSON
#     encoded_payload = json.dumps(payload).encode('utf-8')

#     req = urllib.request.Request(url, data=encoded_payload, headers=headers, method='POST')

#     try:
#         with urllib.request.urlopen(req) as response:
#             if response.getcode() == 204:
#                 print(f"Workflow '{workflow_name}' triggered successfully for branch '{branch}'.")
#             else:
#                 print(f"Failed to trigger workflow. Status code: {response.getcode()}, Response: {response.read()}")
#     except urllib.error.HTTPError as e:
#         print(f"Failed to trigger workflow. HTTPError: {e}")



import cmipld
import json
from pprint import pprint
process = cmipld.JsonLdProcessor()


res = process.get(
    'https://wcrp-cmip.github.io/CMIP6Plus_CVs/project/tables.json')


# pprint(res)
print(json.dumps(res, indent=2))



# from .components import *
from pydantic import BaseModel, field_validator
from cmipld.utils import git
from pydantic import ValidationError
from rich.console import Console
from rich.table import Table

from . import experiment


def field_test(extended_class):
    # dynamic class
    return type(extended_class.__name__, (BaseModel, extended_class), {})


def multi_field_test(args):
    assert isinstance(args, list)
    return type('multi_field_test', (BaseModel, *args), {})


def table2md(table: Table) -> str:
    # Extract headers

    md = f"#### {table.title}\n"
    headers = [
        f"<span style=\'color:{col.style}\'>{col.header}</span>" for col in table.columns]
    md += "| " + " | ".join(headers) + " |\n"
    md += "| " + " | ".join(["---"] * len(headers)) + " |\n"

    # Extract rows
    for row in range(len(table.columns[0]._cells)):
        # Iterate over the row directly
        md += "| " + " | ".join(c._cells[row] for c in table.columns) + " |\n"

    return md


def handle_pydantic_errors(e: ValidationError):
    """Format and display Pydantic errors with rich."""
    console = Console()

    # Create a table to display errors
    table = Table(title="Validation Errors", show_lines=True)
    table.add_column("Field", style="cyan", no_wrap=True)
    table.add_column("Error Message", style="red")
    table.add_column("Error Type", style="yellow")

    # Populate table with error details
    for err in e.errors():
        field = ".".join(str(part)
                         for part in err['loc'])  # Handle nested fields
        message = err['msg']
        err_type = err['type']
        table.add_row(field, message, err_type)

    # Print the table
    console.print(table)
    mdtable = table2md(table)
    # git.update_summary(f"### Validation Errors\n{mdtable}")
    git.update_issue(
        f"### Validation failed:\n {len(e.errors())} issues found.\n{mdtable}")


def run_checks(pydclass, args):
    # Usage in your run function
    try:
        pydclass(**args)  # Your Pydantic validation
    except ValidationError as e:
        handle_pydantic_errors(e)
        git.close_issue(f"Validation failed: {len(e.errors())} issues found.")
    except Exception as e:
        git.close_issue(f"Unexpected error: {str(e)}")



import os
import cmipld
from cmipld.tests.elements.global_import import *
from typing import List, Optional
import asyncio

elementpath = 'organisations/consortia'
elementtype = "consortium"
owners = ['mip-cmor-tables:']

repo_path = cmipld.utils.git.ldpath('')
repourl = cmipld.utils.git.url()
reposhort = urlmap[repourl]

####################################################
# Issue Templates
####################################################
conf = f'''
[{elementtype}]
    Acronym = "CMIP"
    Name = "Coupled Model Intercomparison Project"
    
    [institutions]
        cmip6_acronyms = [
                "CMIP-IPO",
                "WCRP"
            ]
    # nest institutions here, use the cmip acnronyms which they have been registered with.
'''

description = '''
'''

more_info = ''


####################################################
# Validation Functions
####################################################

class DateRange(BaseModel):
    from_: int = Field(..., alias="from")
    phase: str
    to: int


class Institution(BaseModel):
    id: str = Field(..., alias="@id")


class Member(BaseModel):
    type: str = Field(..., alias="@type")
    dates: List[DateRange]
    institution: Institution
    membership_type: str = Field(..., alias="membership-type")


class Validate(BaseModel):
    id: str = Field(..., alias="@id")
    type: str = Field(..., alias="@type")
    changes: Optional[str] = ""
    cmip_acronym: str = Field(..., alias="cmip-acronym")
    members: List[Member]
    name: str
    status: str
    url: Optional[HttpUrl] = None

    @field_validator('id')
    def id(cls, v):
        if not v.startswith(f"{reposhort}:{elementpath}"):
            raise ValueError('must start with the correct path')
        return v

    @field_validator('type')
    def type(cls, v):
        if v != elementtype:
            raise ValueError(f'{v} must be {elementtype} ')
        return v


    # # after alias resolved
    # @model_validator(mode='before')#before
    # def check_keys(cls, values):
    #     return check_all_keys_present(cls, values)
print(os.popen('pwd').read())


async def get_existing():
    inst = await cmipld.quicklook([os.path.abspath('./JSONLD/organisations/institutions/')])
    cons = await cmipld.quicklook([os.path.abspath('./JSONLD/organisations/consortia/')])
    return dict(institutions=inst, consortia=cons)


####################################################
# Element Class
####################################################
class consortium(MIPConfig):
    def __init__(self) -> None:
        self.checks = Validate

    ##### Config to JSONLD #####
    def create_jsonld(self, conf, write=True):

        self.conf = conf
        self.action = conf.get('action', 'new')
        if 'action' in conf:
            del conf['action']

        exist = asyncio.run(get_existing())

        print(exist)

        institutions = dict([[i['cmip_acronym'], i]
                            for i in exist['institutions']])

        error = ''
        inst = {}
        for i in conf['institutions']:
            if i not in institutions:
                error += f'    - Institution [{i}] does not exists in the institutions file. Please add this to proceed.\n'
            else:
                inst[i] = f"{i} [{institutions[i]['identifiers']['ror']} - {institutions[i]['identifiers']['institution_name']}]"
        if error:
            error = '#Error: \n Pausing submission. Please edit the initial config (above) addressing the issues below to try again. \n\n ' + error
            update_issue(issue_number, error)

        print('aaa')

        self.path = ldname(
            f"{repo_path}/{elementpath}/{self.conf['name'].lower()}.json")

        print(self.path)
        self.new_old_action_checks(elementtype)
        print(self.pullname)

        self.json = cmipld.utils.sorted_dict(self.json)

        if self.validate(self.json):
            return self.json
        else:
            return False


def get_template():
    if reposhort not in owners:
        return None
    print('test disabled - to correct. ')
    # test_config(__file__,conf)
    location = repo_path.replace(
        'JSONLD', f".github/ISSUE_TEMPLATE/{elementtype}.md")
    print(f"Saving {elementtype} to {location}")
    return create_template(elementtype, more_info, conf, location)



"""
CMIP6Plus JSON-LD File Extractor and Organizer

This script traverses through a directory structure, processes JSON-LD files, and organizes
them into graph, context, and version categories. It supports parallel execution for faster
processing and allows exclusion of specific directories and files.

Key Features:
- Parallel file reading using ThreadPoolExecutor
- Customizable file and directory exclusion
- Generation of graph and version data
- Minified JSON output for efficient storage
- Base directory specification
- JSON-LD conformance testing
- Override flag to process all non-skipped repos

Dependencies:
- All libraries are from Python's standard library, except:
  - pyld: For JSON-LD conformance testing

Usage:
python cmip6plus_jsonld_extractor.py [--base-dir DIR] [--exclude-dirs DIR1,DIR2] [--exclude-files FILE1,FILE2] [--override]
"""

import argparse
import concurrent.futures
import json
import mmap
import os
import re
import subprocess
from cmipld.locations import rmap, namesplit
from typing import List, Dict, Any, Tuple

from pyld import jsonld

# Constants
SHORTHAND = ""
DEFAULT_SKIP_FILES = ['schema.jsonld', 'graph.jsonld',
                      ".DS_Store", "create.ipynb", "version.jsonld"]
DEFAULT_SKIP_DIRS = ['JSONLD/archive', 'JSONLD/scripts']

graphfile = './compiled/graph_data.json'


def get_repo_info() -> Tuple[str, str, str]:
    """Retrieve repository information and tags."""
    repo = subprocess.getoutput(
        'git remote get-url origin').replace('.git', '/blob/main/JSONLD').strip()
    cv_tag = subprocess.getoutput(
        "curl -s https://api.github.com/repos/WCRP-CMIP/CMIP6Plus_CVs/tags | jq -r '.[0].name'").strip()
    mip_tag = subprocess.getoutput(
        "curl -s https://api.github.com/repos/PCMDI/mip-cmor-tables/tags | jq -r '.[0].name'").strip()
    return repo, cv_tag, mip_tag


REPO, CVTAG, MIPTAG = get_repo_info()


def errprint(*text):
    RED = "\033[91m"
    RESET = "\033[0m"
    print(RED, text, RESET)


def bprint(*text):
    other = "\033[94m"
    RESET = "\033[0m"
    print(other, text, RESET)


def read_json_file_mmap(file: str, directory: str) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Read a JSON file using memory mapping for efficiency."""
    file_path = os.path.join(directory, file)
    with open(file_path, 'r+b') as f:
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            version_data = githistory(file_path, False)
            return json.loads(mm.read().decode('utf-8')), version_data


def read_all_json_files(directory: str, base_dir: str) -> Tuple[str, str]:
    """Read and process all JSON files in a directory."""
    # skipped = [f"{directory}/{f}" for f in DEFAULT_SKIP_FILES]
    try:
        files = [f for f in os.listdir(directory) if f.endswith(
            '.json') and f not in DEFAULT_SKIP_FILES]
        # print(files)
    except FileNotFoundError:
        errprint('<<<   Dir does not exist:', directory)
        return None, None

    if not os.path.exists(f"{directory}/frame.jsonld"):
        bprint('\t >>>   SKIP ', directory)
        return None, None

    try:
        frame = json.load(open(f"{directory}/frame.jsonld", 'r'))
        prefix = frame['@context']['@vocab'].replace('_', '-')
        if prefix[-1] != ':':
            prefix += ':'
        context = frame['@context']
        # print('>>>   PASS ', directory)
    except Exception as e:
        errprint('<<<  Error', e,  directory)
        return None, None

    # files.remove("frame.json")
    all_data = []
    versioning = []

    def add_prefix_to_keys(data, prefix):
        if isinstance(data, dict):
            return {prefix + k if not k.startswith('@') else k: add_prefix_to_keys(v, prefix) for k, v in data.items()}
        elif isinstance(data, list):
            return [add_prefix_to_keys(item, prefix) for item in data]
        else:
            return data

    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_file = {executor.submit(
            read_json_file_mmap, file, directory): file for file in files}

        for future in concurrent.futures.as_completed(future_to_file):
            try:
                data, v = future.result()

                data['@type'] = f"mip:{data['@type']}"
                all_data.append(data)
                versioning.append(v)
            except Exception as exc:
                errprint(
                    f'File {future_to_file[future]} generated an exception: {exc}')

    # write prefix to all_data
    # write context to graph
    # all_data = [{prefix + k if not k.startswith('@') else k: v for k, v in z.items()} for z in all_data]

    all_data = [add_prefix_to_keys(item, prefix) for item in all_data]

    # print(all_data[0])

    print('ADD WALK HERE TO CONTEXT')

    ldroot = directory.removeprefix(base_dir).removeprefix('/./')

    complete_graph = {
        "@id": 'graph:'+context['@vocab'].replace('_', '-'),
        # ldroot.replace('JSONLD/',rmap[namesplit(REPO)]),

        "@type": "mip:graph",
        "@context": context,
        ':url': REPO.removesuffix('JSONLD')+ldroot,
        ":ldroot": ldroot,
        "@graph": all_data,
        ":files": files
    }

    # JSON-LD conformance test

    try:
        jsonld.expand(complete_graph)
        print(f" \t PASS JSON-LD conformance: {directory}")
    except jsonld.JsonLdError as e:
        errprint(f"<<< FAIL JSON-LD conformance test: {directory}: {e}")

    # Write graph and versioning
    gpath = f"{directory}/graph.jsonld"
    json.dump(complete_graph, open(gpath, 'w'),
              indent=4)  # Minified for efficiency
    vpath = f"{directory}/version.jsonld"
    # Minified for efficiency
    json.dump(versioning, open(vpath, 'w'), indent=4)

    return gpath, vpath


def githistory(file: str, rstr: bool = True) -> Dict[str, Any] | str:
    """Retrieve Git history information for a given file."""
    cmd = rf"""git log -n 1 --pretty=format:'{{
        "version:date": "%cd",
        "version:commit": {{
            "hash": "%H",
            "message": "%s",
            "author": {{
                "name": "%an",
                "email": "%ae"
            }}
        }}
    }}' --date=iso-strict -- {file}"""

    rtn = subprocess.getoutput(cmd).strip()
    ncommits = int(subprocess.getoutput(
        f'git log --oneline -- "{file}" | wc -l').strip() or 0)

    update = {}
    if rtn:
        try:
            update = json.loads(rtn)
        except json.JSONDecodeError:
            print(f"Error decoding JSON from git log for {file}: {rtn}")

    if rstr:
        return f'{file.ljust(30)} - {update.get("version:date", "Unknown date")}'
    else:
        return {
            "@id": file.replace(f'JSONLD', REPO),
            "@type": "version",
            "version:file": file,
            "version:release": {"mip-cmor-tables": MIPTAG, "cmip6plus": CVTAG},
            "version:previous_updates": ncommits,
            **update,
            "version:data": {"@id": file.replace("JSONLD/", SHORTHAND).removesuffix('.json')}
        }


def get_directories_to_process(base_dir: str, skip_dirs: set, override: bool) -> set:
    """Get the set of directories to process based on Git status or override flag."""
    if override:
        all_dirs = set()
        for root, dirs, _ in os.walk(base_dir+'/JSONLD/'):
            for d in dirs:
                if d not in skip_dirs:

                    # full_path = os.path.join(root, d)
                    newd = root.replace('../../', '') + '/' + d

                    all_dirs.add(newd.replace('//', '/'))
                # if full_path.startswith(base_dir) and not any(full_path.startswith(i) for i in skip_dirs):
                #     all_dirs.add(full_path.removeprefix(base_dir))

        return all_dirs

    cmd = f'git diff --name-only $(git log -1 --pretty=format:%H -- {graphfile})^ HEAD'
    output = subprocess.getoutput(cmd)

    if not output:
        # Remove graph and version files from the output
        output = re.sub(r'^JSONLD/.+(?:graph|version)\.json.*\n',
                        "", output, flags=re.MULTILINE)

    # Get unique directories that need updating
    return set(re.findall(r'^JSONLD.*(?=/)', output, re.MULTILINE)) - skip_dirs


def main(base_dir: str, exclude_dirs: List[str], exclude_files: List[str], override: bool):
    """
    Main function to orchestrate the JSON-LD file processing.

    :param base_dir: Base directory to start from
    :param exclude_dirs: List of directories to exclude
    :param exclude_files: List of files to exclude
    :param override: Flag to process all non-skipped repos
    """
    skip_files = DEFAULT_SKIP_FILES + exclude_files
    skip_dirs = set(DEFAULT_SKIP_DIRS + exclude_dirs)

    # Get directories to process
    matched = get_directories_to_process(base_dir, skip_dirs, override)

    print(base_dir, matched)
    if not matched:
        if override:
            print('No directories found in the base directory.')
        else:
            print('EXIT: No new files found! Use --override to process all directories.')
        return

    for directory in matched:
        print(directory)
        full_path = os.path.join(base_dir, directory)
        graph_data, _ = read_all_json_files(full_path, base_dir)
        # print(graph_data)

        if not graph_data:
            continue
        # print('<<<', directory)


def init():
    parser = argparse.ArgumentParser(
        description="Process CMIP6Plus JSON-LD files.")
    parser.add_argument("--base-dir", type=str, default="./",
                        help="Base directory to start from")
    parser.add_argument("--exclude-dirs", type=str,
                        help="Comma-separated list of directories to exclude")
    parser.add_argument("--exclude-files", type=str,
                        help="Comma-separated list of files to exclude")

    parser.add_argument("--output-dir", type=str,
                        default="", help="where to save output")

    parser.add_argument("--override", action="store_true",
                        help="Process all non-skipped repos, ignoring Git status")
    args = parser.parse_args()

    # Remove trailing slash for consistency
    base_dir = args.base_dir.rstrip('/')
    exclude_dirs = args.exclude_dirs.split(',') if args.exclude_dirs else []
    exclude_files = args.exclude_files.split(',') if args.exclude_files else []

    main(base_dir, exclude_dirs, exclude_files, args.override)


if __name__ == "__main__":
    init()



from collections import OrderedDict


class DotAccessibleDict:
    def __init__(self, entries):
        self.entries = dict(entries)
        for key in self.entries.keys():
            self.__dict__[key] = self.entries[key]
            if '-' in key:
                self.__dict__[key.replace('-', '_')] = self.entries[key]

    # def __getattr__(self, name):
    #     if name in self.entries:
    #         return self.entries[name]
    #     else:
    #         raise AttributeError(f"'DotAccessibleDict' object has no attribute '{name}'")

    def __str__(self):
        return str(self.entries.keys())


def sorted_json(dct):
    """
    Sort a dictionary by keys and return an OrderedDict.

    Args:
    dct (dict): The input dictionary to be sorted.

    Returns:
    OrderedDict: A new OrderedDict with items sorted by keys.
    """
    if not isinstance(dct, dict):
        return dct

    od = OrderedDict()
    keys = ['@context', 'id', 'type']
    for i in keys:
        if i in dct:
            od[i] = dct[i]
    for key in sorted(dct.keys()):
        if key not in keys:
            od[key] = dct[key]

    return od

    # return sorted((k, v) for k, v in dct.items()))


def sorted_ctx(dct):
    """
    Sort a dictionary by keys and return an OrderedDict.

    Args:
    dct (dict): The input dictionary to be sorted.

    Returns:
    OrderedDict: A new OrderedDict with items sorted by keys.
    """
    assert isinstance(dct, dict)
    assert '@context' in dct

    ctxlist = []

    islist = True
    if not isinstance(dct['@context'], list):
        dct['@context'] = [dct['@context']]
        islist = False

    for dctx in dct['@context']:

        if isinstance(dctx, str):
            ctxlist.append(dctx)
            continue

        ctx = OrderedDict()

        # lddefinitions
        for ck, cv in sorted(dctx.items()):
            if ck[0] == '@':
                ctx[ck] = cv

        # ld objects
        for ck, cv in sorted(dctx.items()):
            if isinstance(cv, str) and cv[0] == '@':
                ctx[ck] = cv

        # prefix
        for ck, cv in sorted(dctx.items()):
            if ck[-1] == ':':
                ctx[ck] = cv

        # others

        # non context items
        for ck, cv in sorted(dctx.items()):
            if '@context' not in cv:
                ctx[ck] = cv

        # context items (links)
        for ck, cv in sorted(dctx.items()):
            if '@context' in cv:
                ctx[ck] = cv

        ctxlist.append(ctx)

    dct['@context'] = ctxlist

    dct['@embed'] = '@always'

    if not islist:
        dct['@context'] = dct['@context'][0]

    return dct



import cmipld
from cmipld.tests.elements.global_import import *
from typing import List, Optional
import asyncio

elementpath = 'organisations/institutions'
elementtype = "institution"
owners = ['mip-cmor-tables:']

repo_path = cmipld.utils.git.ldpath('')
repourl = cmipld.utils.git.url()
reposhort = urlmap[repourl]

####################################################
# Issue Templates
####################################################
conf = f'''
[{elementtype}]
    Acronym = "CMIP-IPO"
    Full_Name = "Coupled Model Intercomparison Project: International Project Office"
    ROR = "000fg4e24"
    
    # only change the item below to "update" if you are submitting a correction. 
    action = "new"
'''

description = '''
'''

more_info = ''


URL_TEMPLATE = 'https://api.ror.org/organizations/{}'


def similarity(name1, name2):
    from difflib import SequenceMatcher

    matcher = SequenceMatcher(None, name1, name2)
    similarity = matcher.ratio() * 100

    return similarity


async def parse_ror_data(ror, cmip_acronym):
    """Parse ROR data and return relevant information."""

    url = URL_TEMPLATE.format(ror)
    ror_data = await CMIPFileUtils().read_file_url(url)
    if ror_data:

        return {
            "@id": ldname(f"mip-cmor-tables:organisations/institutions/{cmip_acronym.lower()}"),
            "@type": elementtype,
            "cmip_acronym": cmip_acronym,
            "ror": ror_data['id'].split('/')[-1],
            "name": ror_data['name'],
            "url": ror_data.get('links', []),
            "established": ror_data.get('established'),
            "kind": ror_data.get('types', [])[0] if ror_data.get('types') else None,
            "labels": [i['label'] for i in ror_data.get('lables', [])],
            "aliases": ror_data.get('aliases', []),
            "acronyms": ror_data.get('acronyms', []),
            "location": {
                "@id": f"mip-cmor-tables:organisations/institutions/location/{ror_data['id'].split('/')[-1]}",
                "@type": "location",
                "@nest": {
                    "lat":  ror_data['addresses'][0].get('lat') if ror_data.get('addresses') else None,
                    "lon":  ror_data['addresses'][0].get('lat') if ror_data.get('addresses') else None,
                    "city": ror_data['addresses'][0].get('city') if ror_data.get('addresses') else None,
                    "country": list(ror_data['country'].values()) if ror_data.get('country') else None
                }
            }
            #  can reverse match consortiums or members from here.

        }
    else:
        return None


####################################################
# Validation Functions
####################################################

class LocationNestModel(BaseModel):
    lat: Optional[float] = None
    lon: Optional[float] = None
    city: Optional[str] = None
    country: Optional[List[str]] = None


class LocationModel(BaseModel):
    id: str = Field(..., alias="@id")
    type: str = Field(..., alias="@type")
    nest: LocationNestModel = Field(..., alias="@nest")


class Validate(BaseModel):
    '''
    Pydantic model for validating element files
    '''

    id: str = Field(..., alias="@id")
    type: str = Field(..., alias="@type")
    cmip_acronym: str
    ror: str
    name: str
    kind: str
    url: Optional[List[HttpUrl]] = None
    established: Optional[int] = None
    org_type: Optional[str] = Field(None, alias="type")
    labels: List[str] = []
    aliases: List[str] = []
    acronyms: List[str] = []
    location: LocationModel

    @field_validator('id')
    def id(cls, v):
        if not v.startswith(f"{reposhort}:{elementpath}"):
            raise ValueError('must start with the correct path')
        return v

    @field_validator('type')
    def type(cls, v):
        if v != elementtype:
            raise ValueError(f'{v} must be {elementtype} ')
        return v

    @field_validator('name')
    def name(cls, v):
        if len(v) < 3 or len(v) > 250:
            raise ValueError('must be at between 3 and 250 characters')
        return v

    @field_validator('acronyms')
    def acronym_on_ror(cls, v, values):
        # values.data.acronyms === v
        return values.data['cmip_acronym'] in v

    # after alias resolved
    @model_validator(mode='before')  # before
    def check_keys(cls, values):
        return check_all_keys_present(cls, values)


####################################################
# Element Class
####################################################
class institution(MIPConfig):

    def __init__(self) -> None:
        # super().__init__()

        self.checks = Validate

    ##### Config to JSONLD #####
    def create_jsonld(self, conf, write=True):

        self.conf = conf
        self.action = conf.get('action', 'new')
        if 'action' in conf:
            del conf['action']

        self.json = asyncio.run(parse_ror_data(conf['ror'], conf['acronym']))

        self.path = ldname(
            f"{repo_path}/{elementpath}/{self.conf['acronym'].lower()}.json")

        # def update_issue_title (what,payload):
        if self.action == 'new':
            if os.path.exists(self.path):
                gitutils.close_issue(
                    '### File Already Exists \n Closing Issue. \n If you meant to update the file, please change the action to "update"')
                raise FileExistsError(f'File Already Exists {self.path}')

            self.pullname = f"Add {elementtype}: {self.json['cmip_acronym']}"
            # cmipld.utils.git.update_issue_title(f"{self.pullname} [{self.json['cmip_acronym']}] - {self.json['name']}")
        else:
            if not os.path.exists(self.path):
                gitutils.close_issue(
                    '### File Does Not Exist \n Closing Issue. \n To add a new file, please change the action to "new"')
                raise FileNotFoundError(
                    f'Missing File - cannot update: {self.path}')

            self.pullname = f"Update {elementtype}:{self.getid}"

        ##########################
        # Check for Similarity in full name
        ##########################
        likeness = similarity(self.json['name'], self.conf['full_name'])
        comment = f"Similarity: {likeness:.2f}%  \n  `[\"{self.json['name']}\" | \"{self.conf['full_name']}\"]` \n"

        if likeness < 55:
            # exit
            comment = f'# Closing Issue <br> {comment} \n Please review and edit the configuration above. If unsure check the ROR ID {URL_TEMPLATE.format(self.json["ror"])}'
            gitutils.close_issue(comment)

        else:
            gitutils.update_issue(f'## Sanity Check \n {comment}', False)

        self.json = cmipld.utils.sorted_dict(self.json)

        if self.validate(self.json):
            return self.json
        else:
            return False


def get_template():
    if reposhort not in owners:
        return None

    print('test disabled - to correct. ')
    # test_config(__file__,conf)

    location = repo_path.replace(
        'JSONLD', f".github/ISSUE_TEMPLATE/{elementtype}.md")

    print(f"Saving {elementtype} to {location}")

    return create_template(elementtype, more_info, conf, location)



import os
import cmipld
from cmipld import processor
from p_tqdm import p_map
from pyld import jsonld
from .class_depends import Depends
from .contexts import get_context

import signal
# Timeout handler


def handler(signum, frame):
    raise TimeoutError("Function timed out!")


# Set up signal handler for timeout
signal.signal(signal.SIGALRM, handler)


class EmbeddedFrame(Depends):
    def __init__(self, url, timeout=15):
        self.url = processor.resolve_prefix(url)

        self.dependencies = processor.depends(url, graph=True)

        self.context_url = processor.contextify(url)

        # expand_context here
        self.context = get_context(self.context_url)

        self.container = []
        self.me = None

        if '@container' in str(self.context):
            print('WARNING: Part of the loaded context contains a "@container" this may mean referenced items become nested in framing.\n Please check output - this is not     corrected for secondary nested items.')

            self.dirpath = os.path.dirname(self.url)
            # filter out direct path.

            # index for the corpus item
            self.me = [i for i, u in enumerate(
                self.dependencies) if self.dirpath in u]
            print(self.me, self.dirpath, self.dependencies)
            assert len(self.me) == 1
            self.me = self.me[0]

            # get the container items
            self.container = [
                k for k, v in self.context.items() if '@container' in v]

        self.urlindex = self.dependencies

        # if self.corpus takes too long
        signal.alarm(len(self.dependencies)*timeout)

        try:
            self.corpus = {'@graph': p_map(jsonld.expand, self.dependencies)}

        except TimeoutError:
            print('Extraction took to long',
                  'extracting files in serial to check for problems.')

            self.corpus = {'@graph': []}
            for i in self.dependencies:
                # if a problem occurs (error) this will tell us which file is responsible.
                print(f'Loading {i}.')
                self.corpus['@graph'].append(jsonld.expand(i))

            print('Serial loading completed sucessfuly. ')

        signal.alarm(0)  # Cancel the alarm

    def frame(self, frame={}):
        if '@context' not in frame:
            frame['@context'] = self.context_url
        if '@embed' not in frame:
            frame['@embed'] = '@always'
        framed = jsonld.frame(self.corpus, frame, options={
                              'embed': '@always', 'extractAllScripts': True, 'expandContest': True})

        # multiresult
        if '@graph' in framed:
            framed = framed['@graph']
        # single result
        elif isinstance(framed, dict):
            framed = [framed]

        # This block of code is iterating over the `framed` data structure and updating its values
        # based on the `container` attribute of the `EmbeddedFrame` class instance.
        ''' 
        When supplying information for framing, additional fields have the same name. This means that all additional information on the fields becoves concatenated. To account for this we overwrite the fields with the optionally provided information. 
        '''

        # overwrite results
        if self.container:
            overwrite = dict([[i['id'], i] for i in jsonld.compact(
                self.corpus['@graph'][self.me], self.context)['@graph']])
            # print(overwrite)
            for id, val in enumerate(framed):

                for cnt in self.container:
                    if cnt not in val:
                        continue
                    item_overwrite = dict([[i['id'], i]
                                          for i in overwrite[val.get('id')][cnt]])
                    for setitem in framed[id][cnt]:
                        setitem.update(item_overwrite[setitem.get('id')])

        return framed

    def iterative_frame(self, frame):

        # get referenced ids.
        independant_ids = [f'{self.dirpath}/{i}' for i in jsonld.frame(
            self.url, {"@explicit": True})['@graph'] if '@type' in i]

        framed = jsonld.frame({"@graph": [id]+self.corpus})



import pytest
import json

'''
!cd ../tests/;
!pytest -v --file-location='/Users/daniel.ellis/WIPwork/CMIP-LD/cmipld/cvs/CV.json' /Users/daniel.ellis/WIPwork/CMIP-LD/cmipld/tests/cvs
'''


def pytest_addoption(parser):
    parser.addoption("--file-location", action="store",
                     default=".CV.json", help="Path to the CV file")


@pytest.fixture(scope='session')
def file_location(request):
    return request.config.getoption("--file-location")


@pytest.fixture(scope='session')
def cv_data(file_location):
    with open(file_location, 'r') as f:
        content = json.load(f)
    return content



#!/usr/bin/env python3

"""
CMIP JSON-LD Graph Processor

This script processes JSON-LD files related to CMIP (Coupled Model Intercomparison Project),
generates a graph, creates frames, links frames, and updates them.

Usage:
    python cmip_graph_processor.py [options] [file_paths...]

Options:
    --graph PATH          Path to write the graph JSON output
    --frame PATH          Path to write the frame JSON output
    --frame-update PATH   Path to write the updated frame JSON output
    --help                Show this help message and exit

Example:
    python cmip_graph_processor.py --graph ./graph.json --frame ./frames.json file1.json file2.json
"""

import cmipld
from cmipld.graph import JSONLDGraphProcessor
import argparse
import asyncio


def parse_args():
    parser = argparse.ArgumentParser(
        description="Process CMIP JSON-LD files and perform graph operations.")
    parser.add_argument("--graph", help="Path to write the graph JSON output")
    parser.add_argument("--frame", help="Path to write the frame JSON output")
    parser.add_argument("--frame-update", action='store_true',
                        help="Path to write the updated frame JSON output")
    parser.add_argument('files', type=str, nargs='+',
                        help='The file(s) to use to generate the network.json')

    args = parser.parse_args()

    if not args.files:
        parser.error("At least one input file is required.")

    return args


async def main(args):
    # Initialize the graph processor
    graph = JSONLDGraphProcessor()

    # Process input files
    await graph.make_graph(args.files)
    print("Graph processing completed.")

    # Write graph if specified
    if args.graph:
        graph.write(location=args.graph)
        print(f"Graph written to {args.graph}")

    # Generate frames and write if specified
    if args.frame:
        core_frame = graph.generate_frames
        linked_frame = graph.link_frames
        graph.filter_frames
        graph.write(location=args.frame, item='frames')
        print(f"Frames written to {args.frame}")

    # Update frames and write if specified
    if args.frame and args.frame_update:
        graph.update_frames()

    return print(f"Processing completed for {graph.prefix}")


def run():
    args = parse_args()
    asyncio.run(main(args))


if __name__ == "__main__":
    run()


'''

reframe --graph ./compiled/graph.json --frame ./compiled/frames.json --frame-update /Users/daniel.ellis/WIPwork/mip-cmor-tables/compiled/graph_data.json /Users/daniel.ellis/WIPwork/CMIP6Plus_CVs/compiled/graph_data.json


'''




# python -m cmipld.tests.elements
import cmipld.tests.elements
from cmipld.utils.git.actions import parse_md, dispatch
import configparser
# from . import activity
import json
import sys
import os
import re


def action():
    issue_body = os.environ.get('ISSUE_BODY')
    print(issue_body)
    conf = parse_md(issue_body)
    print(conf)
    main(conf)


def main(conf):
    issue_number = os.environ.get('ISSUE_NUMBER')
    issue_submitter = os.environ.get(
        'ISSUE_SUBMITTER') or 'automation@wcrp-cmip.org'
    repo = os.environ.get('REPO').replace(
        'https://github.com', 'https://api.github.com/repos')
    token = os.environ.get('GH_TOKEN')
    errors = []

    # process conf
    config = configparser.ConfigParser()
    config.read_string(conf)

    # print(config.sections())

    for section in config.sections():
        try:
            # data
            entry = config[section]
            # library
            section = section.replace('-', "_")
            entrylib = getattr(cmipld.tests.elements, section)
            # class + data
            entryclass = getattr(entrylib, section)()

            # print(entry)

            sucess = entryclass.create_jsonld(entry, write=True)

            # write
            # change branch

            # if sucess:

            #     payload = {
            #         "event_type": 'new_element',
            #         "client_payload": {
            #             "name": entryclass.pullname, # we need this to define the pull request
            #             "issue": issue_number,
            #             "author" : issue_submitter,
            #             "data" : json.dumps(entryclass.json)
            #         }
            #     }

            #     dispatch(token,payload,repo)

        except ModuleNotFoundError as e:
            errors.append(
                f"No such module {section} in cmipld. Please check the congifuration file template.")
            continue

        if errors:
            print(errors)


if __name__ == '__main__':
    print(action())

    # git add one file.
    # git commit with user
    # git push


# conf = '''
# [institution]
#     Acronym = CMIP-IPO
#     Full_Name = CMIP Model Intercomparison Project Office
#     ROR = 000fg4e24

#     # only change the item below to "update" if you are submitting a correction.
#     action = new

# [activity]
#     Name = CMIP
#     Long_Name =  Coupled Model Intercomparison Project
#     URL = https://wcrp-cmip.org

#     # only change the item below to "update" if you are submitting a correction.
#     action = new

# [sub-experiment-id]
#     name =  new-sub-experiment
#     description = A sample sub-experiment id

#     # only change the item below to "update" if you are submitting a correction.
#     action = new
# '''



from ..locations import mapping, matches


class ProcessURL:

    @staticmethod
    def _extract_base_url(url: str) -> str:
        """
        Extract the base URL from a full URL path.

        Args:
            url: The full URL to process

        Returns:
            The base URL containing protocol and domain
        """
        parts = url.split('/')
        return '/'.join(parts[:3])

    @staticmethod
    def id2ctx(id: str) -> str:
        """
        Convert an ID to a context URL.

        Args:
            id: The ID to convert

        Returns:
            The context URL
        """
        return "/".join(id.split("/")[:-1] + ["_context_"])

    @staticmethod
    def resolve_prefix(query):
        if isinstance(query, str) and not query.startswith('http'):
            m = matches.search(query+':')
            if m:
                match = m.group()
                if len(match)-1 == len(query):
                    query = f"{mapping[match]}graph.jsonld"
                else:
                    query = query.replace(match, mapping[match[:-1]])
                print('Substituting prefix:')
                print(match, query)
        return query



import os
import shlex


import json
import tempfile
import io
import urllib.parse
import atexit
import subprocess


def write_dict_to_tempfile(data):
    # Ensure the input is a dictionary
    # if not isinstance(data, dict):
    #     raise ValueError("Input data must be a dictionary")

    # Create a temporary file with a specific suffix for easier access
    temp_file = tempfile.NamedTemporaryFile(
        mode='w', delete=False, suffix=".json")
    try:
        # Write the dictionary as JSON to the file
        json.dump(data, temp_file, indent=4)
        temp_file_name = temp_file.name

        # URL encode the file path to make it safe
        # temp_file_url = urllib.parse.quote(temp_file_name)

        print(
            f"Dictionary written to temporary file at URL-safe path: {temp_file_name}")

        # Register cleanup to delete the temp file on exit

        return temp_file_name
    finally:
        temp_file.close()


def cleanup_temp_file(file_path):
    """Delete the temporary file upon script exit."""
    if os.path.exists(file_path):
        os.remove(file_path)
        print(f"Temporary file {file_path} has been deleted.")


def open_command_line_program(temp):
    # Prompt user for the command to run
    # command = input("Enter the command-line program you want to open (e.g., 'bash', 'python', 'ping google.com'): ")

    # Split the command into executable and arguments
    args = shlex.split(f"jless {temp}")
    program = args[0]

    try:
        # Replace the current process with the command-line program
        os.execvp(program, args)
    except FileNotFoundError:
        print(f"Error: '{program}' not found.")
    except Exception as e:
        print(f"An error occurred while trying to open '{args}': {e}")


def open_jless_with_memory(temp):
    # Create an in-memory file-like object using StringIO
    mem_file = io.StringIO(json.dumps(temp))

    # Use bash to execute the command (bash will read the input from memory)
    # Use subprocess to pipe the in-memory file content to jless
    try:
        # Run `jless` and send the in-memory content via stdin
        # Note: jless will take over the terminal in the foreground
        subprocess.run(['jless', '--mode', 'line'],
                       input=mem_file.getvalue().encode())
    except FileNotFoundError:
        print("Error: jless not found.")
    except Exception as e:
        print(f"An error occurred: {e}")


def interact(data):
    # temp = write_dict_to_tempfile(data)
    #     print(f"temporary file created at {temp}")
    # open_command_line_program(temp)
    open_jless_with_memory(data)
    # atexit.register(cleanup_temp_file, temp)

# if __name__ == "__main__":

    # data = {"key": "value", "number": {"a":123,"b":456}}
    # temp = write_dict_to_tempfile(data)
    # print(temp)
    # open_command_line_program(temp)
    # atexit.register(cleanup_temp_file, temp)



import json
import sys
import re
import os
import argparse
from typing import Any, Dict, List, Union, Set
from functools import lru_cache
from urllib.parse import urljoin
# from ..utils.urltools import valid_url
from pyld import jsonld
from .interactive import open_jless_with_memory
from ..locations import mapping, matches
from .contexts import get_context, generate_context
from p_tqdm import p_map


from .class_loader import Loader
from .class_linknav import LinkNav
# linknav import depends.
# Depends import processurl.


class JsonLdProcessor(Loader, LinkNav):
    """
    A class for processing JSON-LD documents with recursive expansion and ID resolution.

    Features:
    - Recursively expands and resolves @id fields containing URLs
    - Handles document caching
    - Supports compact and expanded document formats
    - Maintains context handling
    - Extracts document dependencies
    """

    def _resolve_ids(self,
                     data: Union[Dict, List],
                     compact: bool = True,
                     depth: int = 60
                     ) -> Union[Dict, List]:
        """
        Recursively resolve @id fields in the document.

        Args:
            data: The data structure to process
            compact: Whether to compact the resulting document

        Returns:
            The processed data structure with resolved IDs
        """
        if not depth:
            return data

        if isinstance(data, dict):
            if '@id' in data and not '@type' in data and data['@id'].startswith('http'):

                # print('!!!',data['@id'])

                try:

                    expanded = self.expand_document(
                        data['@id'],
                        compact=compact,
                        is_nested=True,
                        depth=depth
                    )

                except jsonld.JsonLdError:
                    print('\n WARNING missing id: ', data['@id'])
                    expanded = None
                if expanded:

                    if len(data.keys()) - 1:
                        # we have additional keys
                        id = data['@id']
                        del data['@id']
                        data = jsonld.compact(
                            {**expanded[0], **data}, expanded[0])
                        print(data.keys())
                        # print(data)

                    else:
                        data = expanded[0]

            return {
                key: self._resolve_ids(value, compact, depth)
                for key, value in data.items()
            }

        elif isinstance(data, list):

            if len(data) > 3 and depth < 2:
                # lets try parallel
                def resolve_id(it):
                    return self._resolve_ids(it, compact, depth)

                return p_map(resolve_id, data)
            else:
                return [self._resolve_ids(item, compact, depth) for item in data]

        return data

    def get(self, query, **kwargs):
        query = self.resolve_prefix(query)
        return self.expand_document(query, **kwargs)

    def frame(self, query, frame=None, embed='@always'):
        query = self.resolve_prefix(query)
        if frame is None:
            # use context of the file as a frame
            frame = query
        if '@embed' not in frame:
            frame['@embed'] = embed
        if '@context' not in frame:
            frame['@context'] = mapping
        else:
            frame['@context'] = {**mapping, **frame['@context']}

        return jsonld.frame(query, frame)

    def compact(self, query, ctx=None):

        if isinstance(query, str):
            query = self.resolve_prefix(query)
            if ctx is None:
                print('No context provided, using the context of the file')
                ctx = query

        elif '@context' in query and ctx == None:
            ctx = query['@context']
        elif ctx is None:
            ctx = {}
        return jsonld.compact(query, ctx)

    @lru_cache(maxsize=None)
    def expand_document(self,
                        jsonld_doc: Union[str, Dict],
                        compact: bool = True,
                        expand_ctx: bool = True,
                        expand_links: bool = True,
                        no_ctx: bool = False,
                        as_json: bool = False,
                        pprint: bool = False,
                        depth: int = 2,
                        is_nested: bool = False) -> List[Dict]:
        """
        Expand a JSON-LD document and resolve all referenced URLs.

        Args:
            jsonld_doc: The JSON-LD document to process (URL or dict)
            compact: Whether to compact the final document
            expand_ctx: Whether to expand the context
            expand_links: Whether to expand linked documents
            is_nested: Whether this is a nested expansion

        Returns:
            List of processed documents
        """
        # doc = self._load_document(jsonld_doc) if isinstance(jsonld_doc, str) else jsonld_doc

        # if isinstance(doc['@context'],str):
        #     if not valid_url(doc['@context']):
        #         doc['@context']

        expanded = jsonld.expand(jsonld_doc, options={
                                 'defaultLoader': self.loader})
        depth -= 1

        # mainfile context
        processed = []
        for item in expanded:

            if expand_links:
                processed_item = self._resolve_ids(item, compact, depth).copy()
            else:
                processed_item = item.copy()

            if compact:
                # and not is_nested:
                processed_item = jsonld.compact(
                    processed_item, self.id2ctx(item['@id']))

            if no_ctx and '@context' in processed_item:
                del processed_item['@context']

            processed.append(processed_item)

        if not no_ctx:
            if isinstance(jsonld_doc, str):
                ctx = get_context(jsonld_doc)
            else:
                ctx = jsonld_doc.get('@context', {})

            for item in expanded:
                if '@context' not in item:
                    item['@context'] = ctx

        if pprint:
            from pprint import pprint
            pprint(processed)
        if as_json:
            return json.dumps(processed, indent=4)
        return processed

    @staticmethod
    def EmbeddedFrame(url, timeout=5):
        from .class_embedded import EmbeddedFrame as ef
        return ef(url)



from pydantic import field_validator
import re
import cmipld

owner = 'ror-community'
repo = 'ror-records'

# tag = cmipld.utils.git.get_tags(owner,repo)[0]['name']
# rors = [i['name'].replace('.json','') for i in cmipld.utils.git.get_contents(owner,repo,tag)]

# Regex pattern for ROR key validation
ror_pattern = re.compile(r'^\d{2}\w{5}\d{2}$')


class ror_field:

    ror: str

    @field_validator('ror', mode='after')
    @classmethod
    def ror_length(cls, value):
        assert len(value) == 9, f'ROR key "{value}" must be 9 characters long'
        return value

    @field_validator('ror', mode='after')
    @classmethod
    def ror_validate(cls, value: str) -> str:
        if not isinstance(value, str) or not ror_pattern.fullmatch(value):
            raise ValueError(
                f'Invalid ROR key format: "{value}". Expected format: "https://ror.org/0XXXXXXYY"'
            )
        return value

    # just a test, may need a bit of updating.

    # @field_validator('ror', mode='after')
    # @classmethod
    # def ror_exists(cls, value: str) -> str:
    #     if value not in rors:
    #         raise FileExistsError(f'ROR key "{value}" not found in the ROR records of latest tag: {tag}')
    #     return value



import cmipld
from cmipld.graph import JSONLDGraphProcessor
import argparse
import asyncio


def parse_args():
    parser = argparse.ArgumentParser(
        description='Generates a network.json file from a list of files.')
    parser.add_argument('--output', type=str, default='./network.json',
                        help='The file to write the network.json to.')
    parser.add_argument('--frame', default=False,
                        help='The file to write the network.json to.')
    parser.add_argument('--updateframes', action='store_true',
                        help='Update the frame.jsonls files in each repository.')

    parser.add_argument('files', type=str, nargs='+',
                        help='The file(s) to use to generate the network.json')
    return parser.parse_args()


async def main(args):
    print(args)
    graph = JSONLDGraphProcessor()
    await graph.make_graph(args.files)
    graph.write(location=args.output)

    if args.frame:
        from cmipld.reframe import EnhancedJSONLDGraphProcessor

        frames = graph.generate_frames()
        graph.write(location=args)

    return True


def run():
    args = parse_args()
    asyncio.run(main(args))


if __name__ == "__main__":
    run()


'''

makegraph --output network.json /Users/daniel.ellis/WIPwork/mip-cmor-tables/compiled/graph_data.json /Users/daniel.ellis/WIPwork/CMIP6Plus_CVs/compiled/graph_data.json


'''



import typing
from urllib.parse import urljoin
import requests
import logging


logger = logging.getLogger(__name__)


def generate_context(base='https://wcrp-cmip.github.io', vocab=None, type=None, id=None, expand={}, expand_ctx={}):

    from ..locations import mapping
    from ..utils import sorted_ctx

    context = mapping.copy()

    if base:
        context['@base'] = base
    if vocab:
        context['@vocab'] = vocab

    context.update(expand_ctx)

    context = {"@context": context}

    if type:
        context['@type'] = type
    if id:
        context['@id'] = id

    context.update(expand)

    return sorted_ctx(context)


class ContextResolutionError(Exception):
    """Custom exception for context resolution errors."""
    pass


def resolve_contexts(
    contexts: typing.Union[str, list, dict],
    base_uri: str = '.'
) -> typing.List[dict]:
    """Recursively resolve and expand @context URIs or inline contexts."""
    try:
        if isinstance(contexts, str):
            resolved_uri = urljoin(base_uri, contexts)
            logger.info(f'Resolving context from URI: {resolved_uri}')
            return fetch_resolved_context(resolved_uri)

        if isinstance(contexts, list):
            return [
                ctx_item
                for ctx in contexts
                for ctx_item in resolve_contexts(ctx, base_uri)
            ]

        if isinstance(contexts, dict):
            return [contexts]

        raise ContextResolutionError(
            f"Unsupported @context format: {type(contexts)}")

    except Exception as e:
        logger.error(f"Context resolution error: {e}")
        raise ContextResolutionError(f"Resolution failed: {e}")


def combine_contexts(
    context: typing.Union[list, dict]
) -> dict:
    """Combine multiple context dictionaries into a single dictionary."""
    if isinstance(context, list):
        # context.reverse()
        return {k: v for ctx in context if isinstance(ctx, dict) for k, v in ctx.items()}
        # print(context)
        # ctx = {}
        # (ctx.update(c) for c in context)
        # return ctx

    if isinstance(context, dict):
        return context

    raise ContextResolutionError(f"Invalid context type: {type(context)}")


def fetch_resolved_context(
    jsonld_uri: str,
    timeout: int = 10
) -> typing.List[dict]:
    """Fetch and resolve a JSON-LD document's @context."""
    try:
        response = requests.get(jsonld_uri, timeout=timeout)
        response.raise_for_status()
        context = response.json().get('@context')
        return resolve_contexts(context, jsonld_uri) if context else []

    except requests.RequestException as e:
        logger.error(f"Context fetch failed: {e}")
        raise ContextResolutionError(f"Fetch error: {e}")


def get_context(jsonld_uri: str) -> dict:
    """Retrieve and combine a JSON-LD document's resolved contexts."""
    return combine_contexts(fetch_resolved_context(jsonld_uri))



from cmipld.utils.ldparse import *


def license(data):
    return data['license']


def organisation(data):
    # return 'None if this is missing' Does not break CVs.
    return [name_entry(o['organisation'], 'long_label') if isinstance(o['organisation'], dict) else {o['organisation']: None} for o in data]


functionlist = {}

##################################
# Literals
##################################

literals = ['tracking-id', 'drs', 'index',
            'required_global_attributes', 'repo']

for name in literals:
    field = name.replace('-', '_')

    def literal_wrapper(data, field=field):
        return data[field]
    functionlist[name] = literal_wrapper


##################################
# Name: Description pairs.
##################################

name_description = ['realm', 'frequency', 'grid_label',
                    'nominal', 'sub_experiment', 'tables']

for name in name_description:
    field = name.replace('-', '_')

    def nd_wrapper(data, field=field):
        return sorted(name_entry(data))
    functionlist[name] = nd_wrapper


##################################
# Label Extraction
##################################

'''
All of the following names formulate a function of the format

def <name> (data):
    return get_entry(data[<name>])

This function saves us having to hard code them all. 

# ids are specified with a '-' as not to affect path parsing, 
# however keys are not allowed '-' and must contain '_' instead.
'''

# List of function names
named_entry_feed = ['mip_era', 'activity', 'product', 'tables']
# 'tables',

for field in named_entry_feed:
    functionlist[field] = lambda data, field=field: rmkeys(
        get_entry(data[field], entry='label'))

# save the functions to the local global list
locals().update(functionlist)


# end named field


def experiment(data):
    dummy = {}

    for e in data:

        e['additional_allowed_model_components'] = [
            i for i in e['model_realms'] if not i['is_required']]
        e['required_model_components'] = [
            i for i in e['model_realms'] if i['is_required']]

        # e['model_components_additional_allowed'] = [i for i in e['model_realms'] if not i['is_required']]
        # e['model_components_required'] = [i for i in e['model_realms'] if i['is_required']]

        for name in ['activity', 'parent_experiment', 'additional_allowed_model_components', 'required_model_components']:
            e[name] = name_entry(e[name])

        dummy[e['label']] = rmkeys(e, ['id', 'type', 'model_realms'])

        # 'sub_experiment',

    return dummy


def source(data):

    license_keep = ['label', 'url',
                    'exceptions_contact', 'source_specific_info']
    dummy = {}
    for s in data:
        s['license'] = s['license'][0]
        s['license'] = key_extract(s['license'], license_keep)

        s['organisation'] = name_entry('long_label')
        # {k.split('/')[-1]:v for k,v in s['organisation']},

        s['institution'] = s['organisation']

        s['model_component'] = [key_extract(
            x, ['realm', 'description', 'label', 'native-nominal-resolution']) for x in s['model_component']]

        s['source'] = ''
        for mc in s['model_component']:
            s['source'] += f"{mc['label']} ({mc['realm']}); "

        s = rmkeys(s, ['id', 'type', 'organisation', 'model_component'])

        dummy[s['label']] = s

    return dummy




import json
import sys
import os
import re
import importlib.util


path = '.github/ISSUE_SCRIPT/'


def get_issue():
    return {
        'body': os.environ.get('ISSUE_BODY'),
        "labels_full": os.environ.get('ISSUE_LABELS'),
        'number': os.environ.get('ISSUE_NUMBER'),
        'title': os.environ.get('ISSUE_TITLE'),
        'author': os.environ.get('ISSUE_SUBMITTER')
    }


def parse_issue_body(issue_body):
    lines = issue_body.split('\n')
    issue_data = {}
    current_key = None

    for line in lines:
        if line.startswith('### '):
            current_key = line[4:].strip().replace(' ', '-').lower()
            issue_data[current_key] = ''
        elif current_key:
            issue_data[current_key] += line.strip() + ' '

    # Remove trailing spaces
    for key in issue_data:
        issue_data[key] = issue_data[key].strip()

        if issue_data[key] == "\"none\"":
            issue_data[key] = issue_data[key].replace("\"none\"", "none")

    return issue_data

    # return json.dumps(issue_data, indent=4)


def main():
    issue = get_issue()
    parsed_issue = parse_issue_body(issue['body'])
    issue_type = parsed_issue.get('issue-type', '')
    # print(json.dumps(parsed_issue,indent=4))

    if not issue_type:
        print(json.dumps(parsed_issue, indent=4))
        sys.exit('No issue type selected. Exiting')

    # Define the path to the script based on the issue_type
    script_path = f"{path}{issue_type}.py"

    # Check if the script exists
    if os.path.exists(script_path):
        # Load the script dynamically
        spec = importlib.util.spec_from_file_location(issue_type, script_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        print(f"Successfully loaded {script_path}")
        # run the processing script
        module.run(parsed_issue, issue)

    else:
        print(f"Script {script_path} does not exist")


'''
# GitHub API URL for the issue
url = f'https://api.github.com/repos/{repo_owner}/{repo_name}/issues/{issue_number}'
# https://github.com/WCRP-CMIP/WCRP-universe/issues/2
# Headers for authentication
headers = {
    # 'Authorization': f'token {github_token}',
    'Accept': 'application/vnd.github.v3+json'
}

# Fetch the issue content
response = requests.get(url, headers=headers)
content = response.json()
issue = content.get('body', '')
labels_full = content.get('labels','')

'''



from .__init__ import *
from ..utils import sorted_json


def main():
    description = """CLI entry point for the JSON-LD processor.\n\n
    Process JSON-LD documents and extract dependencies
    
    To toggle between data and line mode, press 'm'. 
    To exit the viewer, press 'q'.
    
    
    Maintainer: Dan Ellis (CMIP IPO)
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('url', help='URL of the JSON-LD document to process')
    parser.add_argument('--deps', '-d', action='store_true',
                        help='Extract dependencies')
    parser.add_argument('--relative', action='store_true',
                        help='Use relative URLs for dependencies')
    parser.add_argument('--no-compact', '-nc',
                        action='store_true', help='Do not compact the document')
    parser.add_argument('--expand-ctx', '-ctx',
                        action='store_false', help='Do not expand context')
    parser.add_argument('--no-expand-links', '-nl',
                        action='store_true', help='Do not expand linked documents')
    parser.add_argument('--output', '-o', help='Output file (default: stdout)')
    parser.add_argument('--no-interactive', '-n',
                        action='store_false',  help='Interactive Playground')
    # parser.add_argument('--no-compact', action='store_true', help='Do not compact the document')
    # data mode for jless

    args = parser.parse_args()
    processor = JsonLdProcessor()

    if 'https://' not in args.url and re.match(matches, args.url):
        for k, v in mapping.items():
            if k in args.url:
                args.url = args.url.replace(k+":", v)
                print(f"Resolving: {k}")
                break
    print('-'*50)
    print(args.url)
    print('-'*50)

    passes = 2
    while passes:
        try:
            if args.deps:

                # Extract dependencies
                deps = processor.extract_dependencies(args.url, args.relative)
                result = sorted(list(deps))
            else:
                # Process document
                result = processor.get(
                    args.url,
                    compact=not args.no_compact,
                    expand_ctx=not args.expand_ctx,
                    expand_links=not args.no_expand_links
                )

                result = sorted_json(result)
            # Output results

            if args.output:
                output = json.dumps(result, indent=2)
                with open(args.output, 'w') as f:
                    f.write(output)

            if args.no_interactive:
                open_jless_with_memory(result)
            else:
                print(output)

            passes = 0
            sys.exit(0)

        except Exception as e:
            if 'Could not retrieve a JSON-LD document from the URL.' in str(e):
                passes -= 1
                args.url += '/graph'
            if not passes:
                print(
                    f"Error processing document: {str(e)}\n\n", file=sys.stderr)
                sys.exit(1)


main()



import os
from .class_depends import Depends


class LinkNav(Depends):
    def find_missing(self, url):
        '''
        Get all the references in an LD object, 
        and check if they exist.
        '''
        from tqdm import tqdm
        from ..utils.urltools import url_exists

        links = self.depends(url)
        missing = [link for link in tqdm(links) if not url_exists(link)]

        return missing

    @staticmethod
    def graphify(path, file='graph.jsonld'):
        return os.path.join(os.path.dirname(path), file)

    def contextify(self, path):
        return self.graphify(path, '_context_')



# import the library

async def quicklook(graphpath, clean=True):
    ''' Quickly load the files from a CMIPLD repo using the frame inside. '''

    import cmipld
    import json

    latest = await cmipld.CMIPFileUtils.load(graphpath)

    print(latest)

    frame = json.load(open(graphpath[0].replace(
        'graph.jsonld', 'frame.jsonld'), 'r'))
    # del frame['@context']
    print(frame)

    return cmipld.Frame([latest], frame).clean().json if clean else cmipld.Frame(latest, frame).json


await quicklook(['JSONLD/organisations/institutions/graph.jsonld'])



'''
Generation script for creating a CV from the CMIP6Plus and MIP tables.

python -m cmipld.cvs.generate

'''
# python -m cmipld.cvs.generate

# Import the library
from cmipld import *
from cmipld.utils.classfn import sorted_dict
from cmipld.utils.git import update_summary, getbranch, getreponame, getlasttag, getlastcommit
import asyncio
import json
import os
import glob
from collections import OrderedDict
from .parse import process
from datetime import datetime


async def main():

    # latest = await sum([mip,cmip6plus],[])
    latest = await CMIPFileUtils.load(['./compiled/graph_data.json', 'mip_cmor_tables'])

    CV = {}
    # OrderedDict()

    ##################################
    ### MIP Entries #####
    ##################################

    # mip entries
    for key in 'source-type frequency realm grid-label nominal-resolution'.split():

        # run the frame.
        frame = get_frame('mip-cmor-tables', key)

        # get results using frame
        data = Frame(latest, frame)

        # any additional processing?
        print(key)
        add_new = await process('mip-cmor-tables', key, data)

        CV[key.replace('-', '_')] = add_new

    ##################################
    ### CMIP6Plus Core #####
    ##################################

    frame = get_frame('cmip6plus', 'descriptors')
    data = Frame(latest, frame, False).clean(
        ['rmld', 'missing', 'untag', 'lower'])

    add_new = await process('cmip6plus', 'descriptors', data, clean=['rmld', 'missing', 'untag', 'lower'])

    CV.update(add_new)

    # ##################################
    # ### CMIP6Plus #####
    # ##################################
    # # organisations
    # # native-nominal-resolution
    for key in 'organisations activity-id sub-experiment-id experiment-id source-id'.split():

        print(key)
        # run the frame.
        frame = get_frame('cmip6plus', key)
        # get results using frame
        data = Frame(latest, frame)

        add_new = await process('cmip6plus', key, data)

        CV[key.replace('-', '_')] = add_new

    CV['institution_id'] = CV['organisations']
    del CV['organisations']

    print('concluding')
    ##################################
    ### fix the file #####
    ##################################

    CV['version_metadata'] = {
        "file_modified": datetime.now().date().isoformat(),
        "CV": {
            "version": getlasttag() or 'version tag read from repo running  - currently not in it. ',
            "git_commit": getlastcommit(),
            "gitbranch": getbranch()},
        "future": 'miptables, checksum, etc'}

    print('above not fatal - version metadata')

    CV = sorted_dict(CV)

    # import pprint
    # pprint.pprint(CV)
    # print(CV)

    branch = ''
    if getbranch() != 'main':
        branch = '_' + getbranch()

    dirname = getreponame().replace('_CVs', '')
    cvloc = os.path.join(os.getcwd(), f'CVs/')
    writelocation = f'{cvloc}{dirname}{branch}_CV.json'

    if branch == '':
        print('on main branch,removing branch files')
        for file in glob.glob(f'{cvloc}*_CV.json'):
            print('removing:', file)
            os.remove(file)

    with open(writelocation, 'w') as f:
        json.dump(dict(CV=CV), f, indent=4)
        print('written to ', f.name)

    update_summary(f'- [x] CV written to {writelocation}')

    return os.path.abspath(writelocation)


def test(writelocation):

    # import pytest

    # # Run pytest and capture the result
    # testsuite =os.path.abspath(os.path.join(os.path.dirname(__file__), '../tests/cvs/'))
    # print(os.getcwd(), __file__)
    # print(testsuite,writelocation)
    # result = pytest.main(["-v",f"--file-location={writelocation}", f"{testsuite}"])

    # print('!!!',result)

    # update_summary(f'CV tests run with exit code {result}')
    result = 'ExitCode.OK'
    update_summary('TESTS SKIPPED - pytest action not finding files. ')

    print('add conditional here')
    if str(result) == 'ExitCode.OK':
        os.popen(f'git add {writelocation}').read()
        os.popen(f'git commit -m "CV generated"').read()

    # # Print a summary based on the result
    # if result == pytest.ExitCode.OK:
    #     print("\nAll tests passed successfully!")
    # elif result == pytest.ExitCode.TESTS_FAILED:
    #     print("\nSome tests failed. Please check the output above for details.")
    # else:
    #     print(f"\nAn error occurred while running the tests. Exit code: {result}")


'''
!cd ../tests/;
!pytest -v --file-location='/Users/daniel.ellis/WIPwork/CMIP-LD/cmipld/cvs/CV.json' /Users/daniel.ellis/WIPwork/CMIP-LD/cmipld/tests/cvs





'''


def run():
    writelocation = asyncio.run(main())
    print('pass cv location into tests')
    test(writelocation)


if __name__ == "__main__":
    run()



import glob,json


files = glob.glob("src-data/*/*.json")

for f in files:
    print(f)
    data = json.load(open(f,'r'))

    rewrite = False
    nd = {}
    for k,v in data.items():
    
        if isinstance(v,list) and len(v) == 1 and isinstance(v[0],dict):
            v = v[0]
            nd[k] = v
            
            rewrite = True
        
        if isinstance(v,dict):
            try:
                # if v.values are all empty dicts 
                if all([i == {} for i in v.values()]):
                    nd[k] = list(v.keys())
                    rewrite = True
                else:
                    nd[k] = v
                    
            except:
                pass
        else:
            nd[k] = v
            

    with open(f,'w') as out:
        if rewrite:
            print(f"Rewriting {f}")
            json.dump(nd,out,indent=2)
            


import requests


def get_repo_tree(owner, repo, path="src-data", branch="main", prefix=None):
    # owner,repo,path,branch = "WCRP-CMIP", "WCRP-universe","src-data","main"

    if prefix is None:
        prefix = repo

    if path[-1] == "/":
        path = path[:-1]
    url = f"https://api.github.com/repos/{owner}/{repo}/contents/{path}?ref={branch}"
    print(url)

    def tree(url):

        # owner,repo = url.replace('https://github.com/','').split('/',2)[:2]
        # preurl = f"https://{owner}.github.io/{repo}/"

        response = requests.get(url)
        contents = response
        if response.status_code == 200:
            contents = response.json()
        else:
            assert not response.text, response

        owner, repo = url.replace(
            'https://api.github.com/repos/', '').split('/', 2)[:2]
        preurl = f"https://{owner}.github.io/{repo}/"

        for item in contents:
            item['prefix'] = prefix

            if '_links' in item:
                del item['_links']
            item['filename'] = ''+item['name']
            item['name'] = item['name'].split('.json')[0]
            item['path'] = item['path'].replace(path, '')
            item['jsonld'] = f"{preurl}{item['path']}".split('.json')[0]
            if item['type'] == "dir":
                item['children'] = tree(item['url'])

        return contents

    contents = tree(url)

    complete = {
        "name": prefix,
        "path": path,
        "prefix": prefix,
        "html_url": f"https://github.com/{owner}/{repo}/blob/{branch}/{path}",
        "git_url": url,
        "children": contents
    }

    return complete


def display_hierarchy(hierarchy, style="arrow", level=0, prefix="", start_level=0, show=False):
    """
    Display a hierarchy in various styles.

    Args:
        hierarchy (list): Hierarchical data (list of dicts with 'name' and optionally 'children').
        style (str): The style of display. Options are:
            - "tree" (default): Tree-like with branches.
            - "indented": Simple indented list.
            - "bracketed": Square bracket notation.
            - "table": ASCII table style.
            - "numbered": Numbered outline format.
            - "graph": Arrows to show relationships.
            - "bullets": Bullet points for levels.
            - "arrow": Downward-right arrow for nesting.
            - "json": JSON-like structure.
        level (int): Current recursion level for indentation (internal use).
        prefix (str): Prefix used for numbering or custom markers (internal use).
        start_level (int): Base level for indentation.

    Returns:
        str: The formatted hierarchy as a string.
    """
    output = ""

    for i, item in enumerate(hierarchy, 1):
        # Determine prefixes and indentation
        current_prefix = f"{prefix}{i}" if style == "numbered" else ""
        indent = "    " * (level - start_level)
        marker = ""

        if style == "tree":
            marker = f"{'│   ' * (level - start_level)}├── {item['name']}"
        elif style == "indented":
            marker = f"{'  ' * (level - start_level)}- {item['name']}"
        elif style == "bracketed":
            marker = f"{'  ' * (level - start_level)}[{item['name']}]"
        elif style == "table":
            marker = f"{indent}{item['name']}"
        elif style == "numbered":
            marker = f"{current_prefix} {item['name']}"
        elif style == "graph":
            marker = f"{'  ' * (level - start_level)}-> {item['name']}"
        elif style == "bullets":
            marker = f"{'*' * (level - start_level + 1)} {item['name']}"
        elif style == "arrow":
            marker = f"{'    ' * (level - start_level)}↳ {item['name']}"
        elif style == "json":
            marker = f"{indent}{{'name': '{item['name']}'"
            if 'children' in item:
                marker += f", 'children': [\n{display_hierarchy(item['children'], style, level + 1, prefix, start_level)}\n{indent}]"
            marker += f"}}"

        # Append the marker to output
        if style != "json":
            output += f"{marker}\n"
        else:
            output += f"{marker}\n"

        # Handle children recursively
        if 'children' in item and style != "json":
            output += display_hierarchy(item['children'], style,
                                        level + 1, current_prefix + ".", start_level)

    if show:
        #     print(output)
        print(output)
    return output


def showall(contents):
    # Display hierarchy with different styles
    print("Tree Style:\n")
    print(display_hierarchy(contents, style="tree"))

    print("\nArrow Style:\n")
    print(display_hierarchy(contents, style="arrow"))

    print("\nIndented List Style:\n")
    print(display_hierarchy(contents, style="indented"))

    print("\nBracketed Style:\n")
    print(display_hierarchy(contents, style="bracketed"))

    print("\nTable Style:\n")
    print(display_hierarchy(contents, style="table"))

    print("\nNumbered Outline Style:\n")
    print(display_hierarchy(contents, style="numbered"))

    print("\nGraph Style:\n")
    print(display_hierarchy(contents, style="graph"))

    print("\nBullet Point Style:\n")
    print(display_hierarchy(contents, style="bullets"))

    print("\nJSON-Like Style:\n")
    print(display_hierarchy(contents, style="json"))


def filter_nested_dict(data, valid_extensions=[".json"],
                       valid_names=["_context_"]):
    """
    Recursively filters a nested dictionary to include only items whose names 
    have valid extensions or match specific names.

    Args:
        data (list): A list of dictionaries representing the nested structure.
        valid_extensions (list): List of valid file extensions (e.g., ['.json', '.jsonld']).
        valid_names (list): List of specific names to include (e.g., ['_context_']).

    Returns:
        list: Filtered nested dictionary.
    """

    def is_valid(name):
        # Check if the name is valid based on extensions or specific names
        return (
            any(name.endswith(ext) for ext in valid_extensions)
            or name in valid_names
        )

    filtered = []
    for item in data['children']:

        if is_valid(item.get('filename', valid_extensions[0])) or item.get('type') != "file":
            # Keep valid items and filter their children recursively
            if 'children' in item:
                item['children'] = filter_nested_dict(
                    item, valid_extensions, valid_names
                )['children']
            filtered.append(item)

    data['children'] = filtered
    return data



from . import functions
from . import add
import os
import subprocess
import sys
import json
import re
from typing import List, Dict, Any, Tuple
import requests
import shlex


def update_env(key, value):
    import os
    github_env_file = os.environ.get("GITHUB_ENV")

    with open(github_env_file, "a") as env_file:
        env_file.write(f"{key}={value}\n")


def update_summary(md):
    import os
    if "GITHUB_STEP_SUMMARY" in os.environ:
        github_env_file = os.environ.get("GITHUB_STEP_SUMMARY")

        with open(github_env_file, "a") as summary:
            summary.write(f"{md}\n")
    else:
        print(md)


def update_issue_title(what):
    # change issue name to reflect contents.
    if 'ISSUE_NUMBER' in os.environ:
        assert isinstance(int(os.environ['ISSUE_NUMBER']), int)
        issue_number = os.environ['ISSUE_NUMBER']

        print(
            os.popen(f'gh issue edit {issue_number} --title "{what}"').read())
        # : {payload["client_payload"]["name"]}"
    update_summary(f"#### Title Updated:\n `{what}`")


def update_issue(comment, err=True, summarize=True):
    if 'ISSUE_NUMBER' in os.environ:
        assert isinstance(int(os.environ['ISSUE_NUMBER']), int)
        issue_number = os.environ['ISSUE_NUMBER']

        cmd = f'gh issue comment {issue_number} --body \'{comment}\' '
        # .replace('"', '\\"').replace('\n', '\\n')
        print(cmd)
        out = os.popen(cmd).read()

        if summarize:
            update_summary(comment)
        if err:
            print(out)
            raise ValueError(comment)

    print(comment)


def close_issue(comment, err=True):
    if 'ISSUE_NUMBER' in os.environ:
        issue_number = os.environ['ISSUE_NUMBER']
        print(os.popen(f'gh issue close {issue_number} -c "{comment}"'))
        if err:
            raise ValueError(comment)
            # sys.exit(comment)

    print(comment)


def jr(file):
    return json.load(open(file, 'r'))


def jw(data, file):
    return json.dump(data, open(file, 'w'), indent=4)


def getfile(fileend):
    import glob
    return glob.glob(f'*{fileend}.json')


def pp(js):
    import pprint
    pprint.pprint(js)


def extract_repo_info(github_pages_url):
    """Extract username and repository name from GitHub Pages URL."""
    pattern = r'https{0,1}://([a-zA-Z0-9-_]+)\.github\.io/([a-zA-Z0-9-_]+)/(.*)?'
    match = re.match(pattern, github_pages_url)

    if match:
        username = match.group(1)
        repo_name = match.group(2)
        path = match.group(3)
        return username, repo_name, path
    else:
        raise ValueError("Invalid GitHub Pages URL")


def io2url(github_pages_url, branch='main', path_base=''):

    username, repo_name, path = extract_repo_info(github_pages_url)
    base_url = f'https://github.com/{username}/{repo_name}/tree/{branch}/{path_base}{path}'

    return base_url


def io2repo(github_pages_url):

    username, repo_name, path = extract_repo_info(github_pages_url)
    base_url = f'https://github.com/{username}/{repo_name}.git'

    return base_url


def url2io(github_repo_url, branch='main', path_base=''):
    print('make test for url2io')

    if '/tree/' in github_repo_url:
        # Regex to extract username, repo name, and path from GitHub repo URL
        pattern = rf"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)/tree/{branch}/{path_base}(?P<path>.*)"

    else:
        pattern = rf"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)"

    match = re.match(pattern, github_repo_url)

    if not match:
        raise ValueError("Invalid GitHub repository URL format.")

    username = match.group("username")
    repo_name = match.group("repo_name")
    path = match.groupdict().get("path", "").strip('/')

    github_pages_url = f"https://{username.lower()}.github.io/{repo_name}/{path}/"
    if github_pages_url[-2:] == '//':
        github_pages_url = github_pages_url[:-1]
    elif github_pages_url[-1] != '/':
        github_pages_url += '/'

    return github_pages_url


def toplevel():
    return os.popen('git rev-parse --show-toplevel').read().strip()


def ldpath(path=''):
    # path = f'organisations/institutions'
    loc = os.path.abspath(f"{toplevel()}/src-data/{path}/")
    if loc[-1] != '/':
        loc += '/'
    return loc


def url():
    return subprocess.getoutput('git remote get-url origin').replace('.git', '').strip()


def commit_override_author(entry, where):
    if os.environ.get('OVERRIDE_AUTHOR'):
        cmd = f'git commit -a --author="{os.environ["OVERRIDE_AUTHOR"].strip()} <{os.environ["OVERRIDE_AUTHOR"].strip()}@users.noreply.github.com>" -m "New entry {entry} to the {where} LD file"'
        print(os.popen(cmd).read())
        return True
    print('No author found')
    os.popen(f'git commit -m "A commit to {entry} at {where}"').read()
    return False


def commit_one(location, author, comment, branch=None):
    cmds = [
        f'git config --global user.email "{author}@users.noreply.github.com"',
        f'git config --global user.name "{author}"',
        f'git add {location} ',
        f'git commit -a --author="{author} <{author}@users.noreply.github.com>" -m "{comment}"'
    ]
    print('>> pushing commit to branch')

    if branch:
        cmds.append(f'git push origin {branch} --force')
    # else:

    cmds.append(f'git push -f ')

    for cmd in cmds:
        print(cmd, ':', subprocess.getoutput(cmd).strip())


def commit(message):
    print(os.popen(f'git commit -a -m "{message}"').read())


def addfile(file):
    print(os.popen(f'git add {file}').read())


def getbranch():
    return subprocess.getoutput('git rev-parse --abbrev-ref HEAD').strip()


def getreponame():
    return subprocess.getoutput('git remote get-url origin').split('/')[-1].replace('.git', '').strip()


def getrepoowner():
    return subprocess.getoutput('git remote get-url origin').split('/')[-2].strip()


def getlastcommit():
    return subprocess.getoutput('git rev-parse HEAD').strip()


def getlasttag():
    return subprocess.getoutput('git describe --tags --abbrev=0').strip()


def getfilenames(branch='main'):
    return os.popen(f'git ls-tree -r {branch} --name-only ').read().split()


def addall():
    print(os.popen('git add -A').read())


def newbranch(branch):
    branch = branch.replace(' ', '-')
    # pull the main
    print(os.popen(f"git pull").read())
    # create the new branch
    print(
        os.popen(f"git checkout -b {branch} || git checkout {branch}").read())
    # pull the branch if it exists
    print(os.popen(f"git pull").read())
    # set upstream
    print(os.popen(f'git branch --set-upstream-to=origin/{branch}').read())


def get_cmip_repo_info() -> Tuple[str, str, str]:
    """Retrieve repository information and tags."""
    repo = subprocess.getoutput(
        'git remote get-url origin').replace('.git', '/blob/main/JSONLD').strip()
    cv_tag = subprocess.getoutput(
        "curl -s https://api.github.com/repos/WCRP-CMIP/CMIP6Plus_CVs/tags | jq -r '.[0].name'").strip()
    mip_tag = subprocess.getoutput(
        "curl -s https://api.github.com/repos/PCMDI/mip-cmor-tables/tags | jq -r '.[0].name'").strip()
    return repo, cv_tag, mip_tag


def branchinfo(feature_branch):
    binfo = subprocess.getoutput(
        f"git rev-parse --verify {feature_branch}").strip()
    if 'fatal' in binfo:
        return False
    return binfo


def reset_branch(feature_branch):
    # if a branch exists, reset it to main, then progress.

    binfo = branchinfo(feature_branch)
    print('BINFO:', binfo)

    cmds = [
        'git remote -v',
        'git fetch --all',
        f"git pull",
        f"git checkout {feature_branch}",
        f"git reset --hard origin/main",
        f"git push origin {feature_branch} -f",


    ]
    if not binfo:
        cmds[3] = f"git checkout -b {feature_branch}"
        cmds[5] = f"git push --set-upstream origin {feature_branch} --force"

    for cmd in cmds:
        print(os.popen(cmd).read())


def prepare_pull(feature_branch):
    issue_number = os.environ['ISSUE_NUMBER']
    if issue_number:
        feature_branch = f'{feature_branch}-{issue_number}'
        reset_branch(feature_branch)
        return feature_branch
    return False


def newpull(feature_branch, author, content, title, issue, base_branch='main', update=None):
    # Ensure we're using the current branch
    feature_branch = subprocess.getoutput("git rev-parse --abbrev-ref HEAD")

    # Set upstream for the current branch if not already set
    os.popen(
        f"git branch --set-upstream-to=origin/{base_branch} {feature_branch}")

    # Ensure there are commits between the branches
    commits = subprocess.getoutput(
        f"git log origin/{base_branch}..HEAD --oneline")
    if not commits:
        raise ValueError(
            f"No commits between {base_branch} and {feature_branch}. Cannot create pull request.")

    # Create the PR or update existing one
    if update is not None:
        where = f"gh pr comment {update}"
    else:
        where = f"gh pr create --base '{base_branch}' --head '{feature_branch}' --title '{title}'"

    # Construct command using cat <<EOF to handle special characters safely
    content = content.replace('`', r'\`')
    cmds = f"""
    nohup git pull -v > /dev/null 2>&1 ;
    {where} --body \"$(cat <<EOF
This pull request was automatically created by a GitHub Actions workflow.


Adding the following new data.

\`\`\`js
{content}
\`\`\`

Resolves #{issue}
EOF
)\"
"""
    # Data submitted by @{author}

    print('++', cmds)

    output = subprocess.getoutput(cmds).strip()

    # Add a comment to the issue if necessary
    update_issue(f'New Pull Request: {output}', False)


def pull_req(feature_branch, author, content, title):
    # gh_token, issue, base_branch
    # Set git configuration

    feature_branch = f'{feature_branch}'

    if not branchinfo(feature_branch):
        print(f'Pull_req: Branch {feature_branch} not found')
        sys.exit(f'Pull_req: Branch {feature_branch} not found')

    cmds = [
        f'git config --global user.email "{author}@users.noreply.github.com" ',
        f'git config --global user.name "{author}" '
        # f'git commit -a --author="{author} <{author}@users.noreply.github.com>" -m "{comment}"'
    ]

    for cmd in cmds:
        print(cmd, ':', subprocess.getoutput(cmd).strip())
    # remote_branch=f'origin/{feature_branch}'
    '''
    gh_token = os.getenv('GH_TOKEN')
    github_repository = os.getenv('GITHUB_REPOSITORY')
    '''
    # feature_branch = os.getenv('FEATURE_BRANCH')

    # Construct the curl command
    # curl_command = (
    #     f'curl -s -H "Authorization: token {gh_token}" '
    #     f'"https://api.github.com/repos/{github_repository}/pulls?state=open&head=origin/{feature_branch}"| jq -r ".[].number"'
    # )

    curl_command = f"gh pr list --head {feature_branch} --state all --json url --jq '.[].url'"

    # Execute the command
    pullrqsts = subprocess.getoutput(curl_command).strip().split('/')[-1]

    # write = False

    # #### sort out existing issue description updates later.

    # # for pr in pullrqsts.split(' '):
    # #     if pr == os.environ["ISSUE_NUMBER"]:
    # #         write = True
    # #         update_issue(f'Overwriting Pull Request Info: |{pr}|',False)
    # #         newpull('main', feature_branch,author,content,title,os.environ["ISSUE_NUMBER"],update = pr)

    # if not write:

    try:
        update = int(pullrqsts)
    except:
        update = None
        # create a new pull request
    newpull('main', feature_branch, author, content,
            title, os.environ["ISSUE_NUMBER"], update)


def push(branch='HEAD'):
    print(os.popen(f'git push -u origin {branch}').read())


def get_tags(owner, repo):
    # Get the tags from the repo
    return requests.get(f'https://api.github.com/repos/{owner}/{repo}/tags').json()


def get_contents(owner, repo, path):
    # Get the tags from the repo
    return requests.get(f'https://api.github.com/repos/{owner}/{repo}/contents/{path}').json()



import os
import sys


def runread(cmd):
    print(os.popen(cmd).read())


runread('git pull')
cmd = r'''
    cd ../../../
    git filter-branch -f --commit-filter \'
	if [ "$GIT_AUTHOR_EMAIL" = "actions@wcrp-cmip.org" ] || [ "$GIT_COMMITTER_EMAIL" = "actions@wcrp-cmip.org" ];
	then
		skip_commit "$@";
	else
		git commit-tree "$@";
	fi\' HEAD
'''
runread(cmd)


def forcepush():
    runread('git push -f origin main')


try:
    if sys.argv[1].strip() == '-f':
        forcepush()
except:
    pass



import http.server
import socketserver

PORT = 8000

Handler = http.server.SimpleHTTPRequestHandler

with socketserver.TCPServer(("", PORT), Handler) as httpd:
    print("Serving at port", PORT)
    httpd.serve_forever()



#!/bin/python3

import glob
import os
import tqdm


def main():

    # do the equivalent of this, but only for the updated repos.

    os.system('update_ctx')

    # os.system('update_schema')

    # os.system('update_issues')

    os.system('validjsonld')

    for i in tqdm.tqdm(glob.glob('src-data/*/')):

        os.system('ld2graph '+i)



import json
import os
import sys
from collections import OrderedDict
import cmipld

# from schema import Schema, And, Use, Optional, SchemaError

from cmipld.locations import urlmap
from cmipld.utils import errprint
import cmipld.utils.git as gitutils

from cmipld.file_io import CMIPFileUtils, ldname
from cmipld.frame_ld import Frame

from pydantic import BaseModel, ValidationError, field_validator, Field, HttpUrl, model_validator


def getid(self):
    return self.json['@id'].split('/')[-1]


def pydantic_eprint(e):
    error_details = e.errors()

    gh = '''# Errors <br>'''

    print("\nParsed error details:")
    for error in error_details:
        # print(f"Error in field '{error}")

        gh += f"-  {error['msg']} \n Input:{error['input']} \n\n"

    gitutils.update_issue(f'#Sanity Check \n {comment}')


class MIPConfig():
    '''Superclass for all MIP elements'''
    # def __init__(self) -> None:
    #     self.checks = lambda x: print('checks not set up')

    def validate_one(self, data):
        try:
            data['skip'] = 'all'
            self.checks(**data)
            return True
        except ValidationError as e:
            pydantic_eprint(e)
            return False

    def validate(self, data):
        try:
            self.checks(**data)
            return True
        except ValidationError as e:
            pydantic_eprint(e)
            return False

    @property
    def getid(self):
        return self.json['@id'].split('/')[-1]

    @property
    def jsonstr(self):
        return json.dumps(self.json, indent=4)

    def new_old_action_checks(self, elementtype):
        if self.action == 'new':
            if os.path.exists(self.path):
                gitutils.close_issue(
                    '### File Already Exists \n Closing Issue. \n If you meant to update the file, please change the action to "update"')
                raise FileExistsError(f'File Already Exists {self.path}')

            self.pullname = f"Add {elementtype}: {self.json['cmip_acronym']}"
        else:
            if not os.path.exists(self.path):
                gitutils.close_issue(
                    '### File Does Not Exist \n Closing Issue. \n To add a new file, please change the action to "new"')
                raise FileNotFoundError(
                    f'Missing File - cannot update: {self.path}')

            self.pullname = f"Update {elementtype}:{self.getid}"


def check_all_keys_present(cls, values):
    """
    Ensure that all required fields are present in the input data.
    """
    # required_fields = cls.__annotations__.keys()
    required_fields = {field.alias or name: name for name,
                       field in cls.__fields__.items()}

    # print('-----')
    # for i in cls.__fields__.items():
    #     print(i)
    #     print(i[1].alias)
    #     print(values)
    # print('-----')

    if 'skip' in values:
        if values['skip'] == 'all':
            return values
        else:
            for i in values['skip']:
                required_fields.remove(i)

    # Check for missing fields
    missing_fields = [
        field for field in required_fields if field not in values and f"@{field}" not in values]

    if missing_fields:
        raise ValueError(
            f"Missing required fields: {', '.join(missing_fields)}\n\n")

    return values


def create_template(element_type, moreinfo, config, location):

    element_type_c = element_type.capitalize()
    split = ' '.join([i.capitalize() for i in element_type.split('-')])

    template = f"""---
name: {element_type_c}
about: Adding a new, or updating an existing, {split.lower()}
title: 'Review request for change in {element_type}'
labels: '{element_type}'
assignees: ''

---

# Add Consortium Template

To request a new item, please amend the template below to reflect the items you are interested in.

Relevant conditions and naming conventions for this item can be found using the wiki pages [here]({moreinfo}).

## Amending Information on an Existing Item

If you wish to amend an item, please supply only the fields you are interested in and ensure that you change the *action* field to *update*.

``` action = update ```

<!---  info 

We are trialing the addition of new components using the configuration file format. 
To use this, please fill out the template below, keeping the spacing and indentation of the file.

--->

## Contents (What We Wish to Add)



``` configfile

{config}

```

"""

    with open(location, 'w') as f:
        f.write(template)
        print(f"Template written to {location}")
    return template


def test_config(file, conf):
    # test_config()
    import configparser
    cfg = configparser.ConfigParser()
    cfg.read_string(conf)

    element = __file__.split('/')[-1].split('.')[0]
    templateconf = dict(cfg[element])

    test = globals()[element]()
    test.create_jsonld(templateconf, write=False)



import requests
# import urlparse
from urllib.parse import urlparse, urlunparse
from os.path import relpath


def url_exists(url):
    try:
        response = requests.head(url, allow_redirects=True, timeout=5)
        return response.status_code == 200
    except requests.ConnectionError:
        return None
    except requests.Timeout:
        return None
    except requests.RequestException as e:
        print(f"An error occurred: {e}")
        return None


def relative_url(base_url, target_url):
    base_parsed = urlparse(base_url)
    target_parsed = urlparse(target_url)

    if base_parsed.netloc != target_parsed.netloc:
        raise ValueError(
            "URLs have different domains; cannot compute relative path.")

    # Compute relative path
    base_path = base_parsed.path
    target_path = target_parsed.path
    relative_path = relpath(target_path, start=base_path)

    # Reconstruct the relative URL
    return relative_path


# def valid_url(url):
#     parsed_url = urlparse(url)
#     # Check if the URL has a valid scheme (e.g., 'http', 'https') and netloc (domain)
#     return parsed_url.scheme and parsed_url.netloc


def https(url):
    if url.startswith("http://"):
        return url.replace("http://", "https://", 1)
    return url



import os
import shutil
import subprocess
import tempfile
from typing import List, Tuple
import tarfile
import datetime
# from .. import locations
from .server import LocalServer
from ..locations import reverse_mapping
from .git import io2repo

# loader in read, part of cmipld.processor


# import loader class


class LD_server:
    def __init__(self, repos=None, zipfile=None, copy=None, override=None):
        '''
        If None provided all the cmipld repositories will be generated. 

        repos: List of tuples where each tuple is (repo_url, target_name).
        zipfile: Path to the tar.gz file.

        '''
        self.temp_dir = None
        self.create_temp_dir()

        # ignore
        # if not repos and not zipfile:
        #     print('No repositories or zip file provided. Generating all repositories')
        #     repos = location.reverse_mapping()

        if zipfile:
            self.from_zip(zipfile, override=override)
        if repos:
            self.clone_repos(repos, override=override)
        if copy:
            self.copy_existing_repos(copy, override=override)

    def create_temp_dir(self):
        """Create a temporary directory to hold repositories."""
        if not self.temp_dir:
            self.temp_dir = tempfile.TemporaryDirectory(
                prefix='cmipld_local_', suffix=datetime.datetime.now().isoformat().split('.')[0])
        return self.temp_dir.name

    def delete_temp_dir(self):
        """Delete the temporary directory."""
        if self.temp_dir:
            self.temp_dir.cleanup()
            self.temp_dir = None

    def clone_repos(self, repos: List[Tuple[str, str]], branch="production", dir='src-data', override='n'):
        """
        Clone a list of git repositories into the temporary directory.
        Args:
            repos: List of tuples where each tuple is (repo_url, target_name).
            branch: Branch to clone (default: 'production').
        """
        for repo_url, target_name in repos:
            print(repo_url, target_name, self.temp_dir.name)
            repo_path = os.path.join(self.temp_dir.name, target_name)

            if '.io' in repo_url:
                repo_url = io2repo(repo_url)

            if os.path.exists(repo_path):
                if override != 'y':
                    override = input(
                        f"Repo '{target_name}' already exists. Delete and replace? (y/n): ").lower()

                if override == 'y':
                    shutil.rmtree(repo_path)
                else:
                    print(f'Repo {target_name} not replaced')
                    continue

            clone = os.popen(' '.join(
                ["git", "clone", "--branch", branch, "--single-branch", repo_url, repo_path])).read()
            print(clone)

            assert 'fatal' not in clone

            # move the relevant repo into place. This is because our production branch serves only the src-data directory
            print(os.popen(f'mv {repo_path}/{dir}/* {repo_path}').read())

        print(f"Repositories cloned into {self.temp_dir}")

    def copy_existing_repos(self, repo_paths: List[str], override='n'):
        """
        Copy existing repositories into the temporary directory.
        Args:
            repo_paths: List of file paths to existing repositories.

        E.g. [[path1,name1],[path2,name2]]
        """
        for tocopy in repo_paths:
            if len(tocopy) == 2:
                repo_path, repo_name = tocopy
            else:
                repo_path = tocopy
                repo_name = tocopy

            print('Copying the repo into LocalServer ',
                  repo_path, '-->', repo_name)
            target_name = os.path.basename(repo_name)
            target_path = os.path.join(self.temp_dir.name, target_name)

            if os.path.exists(target_path):
                if override != 'y':
                    override = input(
                        f"Repo '{target_name}' already exists. Delete and replace? (y/n): ").lower()

                if override == 'y':
                    shutil.rmtree(target_path)
                else:
                    continue
            shutil.copytree(repo_path, target_path)

        print(f"Repositories copied into {self.temp_dir}")

    def rollback_repo(self, repo_name: str, commit_hash: str):
        """
        Roll back a repository to a specific commit.
        Args:
            repo_name: Name of the repository to roll back.
            commit_hash: Commit hash to roll back to.
        """
        temp_dir = self.create_temp_dir()
        repo_path = os.path.join(temp_dir.name, repo_name)

        if not os.path.exists(repo_path):
            raise FileNotFoundError(
                f"Repository '{repo_name}' not found in {temp_dir}")

        subprocess.run(["git", "checkout", commit_hash],
                       cwd=repo_path, check=True)
        print(f"Repository '{repo_name}' rolled back to commit {commit_hash}")

    def to_zip(self, output_file: str):
        """
        Create a gzipped tarball of the temporary directory.
        Args:
            output_file: Output tar.gz file path.
        """
        temp_dir = self.create_temp_dir()
        with tarfile.open(output_file, "w:gz") as tar:
            tar.add(temp_dir, arcname=os.path.basename(temp_dir.name))
        print(f"Repositories compressed into {output_file}")

    def from_zip(self, zip_path: str):
        """
        Extract repositories from a gzipped tarball.
        Args:
            zip_path: Path to the tar.gz file.
        """
        temp_dir = self.create_temp_dir()
        with tarfile.open(zip_path, "r:gz") as tar:
            tar.extractall(temp_dir.name)
        print(f"Repositories extracted into {temp_dir}")

    def start_server(self, port=8080):
        '''
        Serve the directory at the specified port.
        '''
        for _ in range(9):
            try:
                self.server = LocalServer(self.temp_dir.name, port)
                break
            except:
                port += 1
                print('Port in use, trying:', port)

        self.url = self.server.start_server()
        return self.url

    def stop_server(self):
        self.server.stop_server()
        self.server = None
        self.url = None
        print("Server stopped.")




from typing import Any, Dict, List, Union, Set
from ..utils.urltools import https
from pyld import jsonld

custom_cache = {}


class Loader:

    def __init__(self, tries=3):
        """Initialize the processor with a cached document loader."""
        self.loader = None
        self.set_cache_loader(tries)
        assert self.loader == jsonld.get_document_loader()

    # @lru_cache(maxsize=100)

    def _load_document(self, url: str) -> Dict:
        """
        Load and cache a JSON-LD document from a URL.

        Args:
            url: The URL to fetch the document from

        Returns:
            The loaded document
        """
        return self.loader(url)['document']

    @staticmethod
    def clear_cache():
        global custom_cache
        custom_cache = {}

    def set_cache_loader(self, tries=3):

        default_loader = jsonld.requests_document_loader()

        def cached_loader(url, kwargs={}):
            global custom_cache
            url = https(url)
            # cache hit
            if url in custom_cache:
                return custom_cache[url]

            # cache miss
            for _ in range(tries):
                try:
                    custom_cache[url] = default_loader(url)
                    return custom_cache[url]
                except:
                    pass

            # last time to throw the error

            custom_cache[url] = default_loader(url)
            return custom_cache[url]

        # update jsonld loader
        jsonld.set_document_loader(cached_loader)

        self.loader = jsonld.get_document_loader()

    def replace_loader(self, server_url, replaced=[], tries=3):
        '''
        replaced = [('https://wcrp-cmip.github.io/CF/', 'cf'),]

        '''
        import warnings
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        warnings.simplefilter('ignore', InsecureRequestWarning)

        assert type(replaced) == list
        self.clear_cache()

        default_loader = jsonld.requests_document_loader(verify=False)

        print('Setting up location forwarding for:')
        for j, i in replaced:
            print(f" -  {j} >>> {server_url}/{i}/ \n")

        def localhost_loader(url, kwargs):
            global custom_cache

            url = https(url)
            if url in custom_cache:
                return custom_cache[url]

            for j, i in replaced:
                if j in url:
                    url = url.replace(j, f'{server_url}/{i}/')
                    break

            for _ in range(tries):
                try:
                    custom_cache[url] = default_loader(url)
                    return custom_cache[url]
                except:
                    pass
            # last time to throw the error
            custom_cache[url] = default_loader(url)
            return custom_cache[url]

        # update jsonld loader
        jsonld.set_document_loader(localhost_loader)

        # update default cmipld loader
        self.loader = jsonld.get_document_loader()



import argparse


def run():
    parser = argparse.ArgumentParser(description="Test inputs")
    parser.add_argument("--flag", action="store_true", help="Flag to test")
    parser.add_argument("--option", help="Option to test")
    parser.add_argument("--updated", nargs="+", help="Inputs to test")

    args = parser.parse_args()
    print('Inputs test')
    print(args)
    return None




from collections import OrderedDict
import cmipld

# def name_description(data,key='name',value='description'):
#     return dict([(x[key],x[value]) for x in data])

# def key_only(data,key='name',default='missing'):
#     if isinstance(data[0],str):
#         # if single value do not fetch key
#         return data
#     return sorted(list(set([x.get(key,default) for x in data if x])))


##################################
### MIP Table fns      ###########
##################################

def mip_cmor_tables_source_type(data):
    return data.key_only('name')


def mip_cmor_tables_frequency(data):
    return data.key_value('name', 'description')


def mip_cmor_tables_realm(data):
    return data.key_value('name', 'description')


def mip_cmor_tables_grid_label(data):
    return data.key_value('name', 'description')


def cmip6plus_activity_id(data):
    return data.key_value('name', 'description')


def cmip6plus_sub_experiment_id(data):
    return data.key_value('name', 'description')


##################################
### CMIP fns            ###########
##################################

def cmip6plus_organisations(data):

    # data = [d['organisation_id'] for d in data if d['organisation_id']['cmip_acronym']]

    data.clean(['rmnull'])

    # # return filter(lambda x: 'cmip_acronym' in x, data)
    # data.print

    return cmipld.key_value(data.value_only('organisation_id'), key='cmip_acronym', value='name')


def cmip6plus_descriptors(data):
    data = data.json

    for i in data['index']:
        if isinstance(data['index'][i], str):
            data['index'][i] = [data['index'][i]]

    for i in 'tracking_id license'.split():
        if isinstance(data[i], str):
            data[i] = [data[i]]

    data.update(data['index'])
    del data['index']
    data['DRS'] = data['drs']
    data['Conventions'] = data['conventions']
    print(data['mip_era'])
    data['mip_era'] = data['mip_era']['name']
    data['product'] = data['product']['kind']

    del data['drs'], data['conventions'], data['@context']
    return data


def cmip6plus_source_id(data):
    data = data.json

    sid = OrderedDict()
    for source in sorted(data, key=lambda x: x['source_id']):
        # ideally organisation
        source['institution_id'] = [
            source['organisation_id'].get('cmip_acronym', '')]
        del source['organisation_id']

        source["activity_participation"] = cmipld.value_only(
            source.get("activity_participation", []), 'name')

        source['license'].update(source['license'].get('kind', {}))

        if not isinstance(source['cohort'], list):
            source['cohort'] = [source['cohort']]

        source['source'] = f"{source['source_id']} ({source['release_year']}): \n  "

        #    combine the model-components
        for i in source['model_component']:
            try:
                source['source'] += f"{i['name']} ({i['realm']['name']})\n  "
            except Exception as e:

                print('Missing', i, source['source_id'])

                print(e)

        # del source['license']['kind']
        # del source['license']['conditions']

        del source['model_component']
        sid[source['source_id']] = source

    return sid


def cmip6plus_native_nominal_resolution(data):
    data = data.json

    return list(set([f"{x['nominal_resolution'].get('value',x['nominal_resolution'])}{x['nominal_resolution'].get('unit',{}).get('si','km')}" for x in data]))


def cmip6plus_experiment_id(data):
    data = data.json

    eid = OrderedDict()
    for e in sorted(data, key=lambda x: x['experiment_id']):

        for i in ['additional_allowed', 'required']:
            if isinstance(e['model_components'][i], str):
                e['model_components'][i] = [e['model_components'][i]]

            e[f'{i}_model_components'] = e['model_components'][i]

        del e['model_components']

        # to list
        e['activity_id'] = cmipld.key_only([e['activity_id']])

        for i in e['parent']:
            e['parent_'+i] = [e['parent'][i]]

        e['sub_experiment_id'] = [
            e['sub_experiment_id'].get('name', 'missing')]

        del e['parent']

        eid[e['experiment_id']] = e

    return eid


##################################
### Main processing function #####
##################################
local_globals = globals()


async def process(prefix, file, data=None, clean=None):
    name = f'{prefix}_{file}'.replace('-', '_')
    # prepare for use.

    if clean:
        # print('clean', clean)
        data.clean(clean)
    else:
        data.clean_cv
    if name in local_globals:
        data.data = local_globals[name](data)
    else:
        print('no parsing function found', name)

    return data.json




from jsonschema import validate, exceptions
from jsonschema import RefResolver

import json
import yaml

# validate(data, schema)


def expand_schema(schema):
    """Expand the schema by resolving all $ref references using a self-contained URI"""
    resolver = RefResolver.from_schema(schema)
    return resolve_ref(schema, resolver)

# expand_schema(schema)


# def convert_schema_to_github_form(schema_path, output_path):

#     # with open(schema_path, 'r') as f:
#     #     schema = json.load(f)

#     schema = schema_path
#     form = {
#         "name": schema.get("title", "Form"),
#         "description": "Auto-generated form from JSON Schema.",
#         "title":"{{ name }} - Submission",
#         "body": []
#     }

#     for field, details in schema.get("properties", {}).items():
#         form_field = {
#             "id": field,
#             "attributes": {
#                 "label": details.get("title", field.replace("_", " ").title()),
#                 "description": details.get("description", ""),
#                 "placeholder": details.get("default", "")

#             },
#     "validations":{"required": field in schema.get("required", [])}
#         }

#         if details["type"] == "string" and "enum" in details:
#             form_field["type"] = "dropdown"
#             form_field["attributes"]["options"] = details["enum"]
#         elif details["type"] == "array" and "items" in details and "enum" in details["items"]:
#             form_field["type"] = "checkboxes"
#             form_field["attributes"]["options"] = [{"label": opt.lower()} for opt in details["items"]["enum"]]
#             form_field['multiple'] = details.get("multiple", False)
#         elif details["type"] == "boolean":
#             form_field["type"] = "dropdown"
#             form_field["attributes"]["options"] = ["Yes", "No"]
#             form_field['miultiple'] = details.get("multiple", False)
#         else:
#             form_field["type"] = "input"
#             form_field["attributes"]["placeholder"] = details.get("default",f"Enter your {field.replace('_', ' ')}")

#         form["body"].append(form_field)

# # checkbox options can be required much like the field.


#     # with open(output_path, 'w') as f:
#     #     yaml.dump(form, f, sort_keys=False)
#     return yaml.dump(form)

# Example Usage



'''
The add script used for dispatch events. 
'''


import json
import sys
import os
import re

# Add the current directory to the Python path
# current_dir = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(current_dir)

# Get the parent directory of the current file
# parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
# sys.path.append(parent_dir)


def submit_dispatch():

    from cmipld.git.actions.functions import parse_md, dispatch, update_issue_title

    # generic items to read the issue
    issue_number = os.environ.get('ISSUE_NUMBER')
    issue_title = os.environ.get('ISSUE_TITLE')
    issue_body = os.environ.get('ISSUE_BODY')
    issue_submitter = os.environ.get('ISSUE_SUBMITTER')
    repo = os.environ.get('REPO').replace(
        'https://github.com', 'https://api.github.com/repos')
    token = os.environ.get('GH_TOKEN')

    #  get content.
    parsed = parse_md(issue_body)

    '''
    Lets submit the data to a dispatch event
    '''

    for kind in parsed:
        print(kind)
        data = parsed[kind]

        payload = {
            "event_type": kind,
            "client_payload": {
                # we need this to define the pull request
                "name": data['acronym'],
                "issue": issue_number,
                "author": issue_submitter,
                "data": json.dumps(data)
            }
        }

        update_issue_title(issue_number, kind, payload)

        dispatch(token, payload, repo)



'''
Remove an already published tag from a repo. 

Author: daniel.ellis@wcrp-cmip.org
'''

import os
import sys

tag = sys.argv[1].strip()


def runread(cmd):
    print(os.popen(cmd).read())


runread(f'git tag -d {tag}')
runread(f'git push origin :refs/tags/{tag}')



"""
Tools for adding and validating checksums
"""
from copy import copy, deepcopy
import hashlib
import json
import datetime
from collections import OrderedDict
from .git import getbranch, getreponame, getrepoowner, getlastcommit, getlasttag, url, url2io
from ..locations import reverse_mapping

# def calculate_checksum(dictionary, overwrite=True, checksum_location='version_metadata'):
#     """
#     Calculate the checksum for dictionary and add it to the Header

#     Parameters
#     ----------
#     dictionary: dict
#         The dictionary to set the checksum for.
#     overwrite: bool
#         Overwrite the existing checksum (default True).
#     checksum_location: str
#         sub-dictionary to look for in /add the checksum to.

#     Raises
#     ------
#     RuntimeError
#         If the ``checksum`` key already exists and ``overwrite`` is
#         False.
#     """
#     if 'checksum' in dictionary['Header'][checksum_location]:
#         if not overwrite:
#             raise RuntimeError('Checksum already exists.')
#         del dictionary['Header'][checksum_location]['checksum']
#     checksum = _checksum(dictionary)
#         dictionary['Header'][checksum_location]['checksum'] = checksum


def validate_checksum(dictionary, checksum_location='version_metadata'):
    """
    Validate the checksum in the ``dictionary``.

    Parameters
    ----------
    dictionary: dict
        The dictionary containing the ``checksum`` to validate.
    checksum_location: str
        sub-dictionary to look for in /add the checksum to.

    Raises
    ------
    KeyError
        If the ``checksum`` key does not exist.
    RuntimeError
        If the ``checksum`` value is invalid.
    """
    if 'checksum' not in dictionary[checksum_location]:
        raise KeyError('No checksum to validate')
    dictionary_copy = deepcopy(dictionary)
    del dictionary_copy[checksum_location]['checksum']
    checksum = _checksum(dictionary_copy)
    if dictionary[checksum_location]['checksum'] != checksum:
        msg = ('Expected checksum   "{}"\n'
               'Calculated checksum "{}"').format(dictionary[checksum_location]['checksum'],
                                                  checksum)
        raise RuntimeError(msg)


def _checksum(obj):
    obj_str = json.dumps(obj, sort_keys=True)
    checksum_hex = hashlib.md5(obj_str.encode(
        'utf8')).hexdigest().split('.')[0]
    return 'md5: {}'.format(checksum_hex)


def version(data, name, location='./', repo=None):
    rmap = reverse_mapping()

    writefile = f'{location}'
    output = OrderedDict()

    header = OrderedDict()
    header['file'] = writefile
    header['file_creation_date'] = datetime.datetime.now().isoformat()
    # header['branch'] = getbranch()
    header['version'] = getlasttag()
    header['checksum'] = _checksum(data)

    if repo:
        header['repo_url'] = repo[0]
        header['repo_prefix'] = repo[1]
    else:
        header['repo_url'] = url()
        header['repo_prefix'] = rmap[url2io(header['repo_url'])]

    header['last_commit'] = getlastcommit()
    header['comment'] = 'This is an automatically generated file. Do not edit.'

    output['Header'] = header
    output[name] = OrderedDict(sorted(data.items(), key=lambda item: item[0]))

    return output



from setuptools import setup, find_packages

setup(
    name="cmip-ld",
    author="Daniel Ellis",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "requests",
        "pyld",
        "jmespath",
        "pytest",
        "pydantic",
        "p_tqdm"
    ],
    entry_points={
        "console_scripts": [
            "cmipld=cmipld.browse.__main__:main",
            "update_ctx=cmipld.generate.update_ctx:main",
            "update_all=cmipld.generate.update_all:main",
            "update_schema=cmipld.schema.update_schema:main",
            "update_issues=cmipld.schema.issue_template:main",
            "rewrite_type=cmipld.generate.rewritetype:main",
            "new_issue=cmipld.generate.new_issue:main"
            # "cmipgraph=cmipld.graph:main",
            # "updateld=cmipld.generate.update_new:init",
            # "makegraph=cmipld.generate.makegraph:run",
            # "reframe=cmipld.generate.reframe:run",
            # "updateelastic=cmipld.generate.elasticmipcvs:run",
            # "add_new=cmipld.git.actions.add:submit_dispatch",
            # "test_inputs=cmipld.tests.inputs:run",
            # "issue_templates=cmipld.generate.issue_template:run",
            # "new_element=cmipld.generate.new_element_from_issue:action",
            # "gencv=cmipld.cvs.generate:run",
        ],
    },
    scripts=[
        # "scripts/directory-utilities/combine-graphs",
        # "scripts/directory-utilities/compile-ld",
        "scripts/directory-utilities/dev",
        "scripts/directory-utilities/validjsonld",
        "scripts/directory-utilities/ld2graph",
        "scripts/jsonld-util/rmbak",
        "scripts/jsonld-util/rmgraph"
        
    ],
    include_package_data=True,
    package_data={
        "cmipld": ["scripts/*/*.sh"],
        'cmipld.frame_ld': ['examples/*/*.json'],
    },
    # package_data={
    #     'your_library': ['examples/*.json'],  # Adjust this to your actual package and files
    # },
)



import glob
import json 
from collections import OrderedDict
import os
import subprocess





remote_url = subprocess.check_output(["git", "remote", "get-url", "origin"], text=True).strip()
user_repo = remote_url.split("github.com")[-1].replace(":", "/").replace(".git", "").strip("/")
pages_url = f"https://{user_repo.split('/')[0]}.github.io/{user_repo.split('/')[1]}"

# print(remote_url)
# print(user_repo)
# print(pages_url)
# https://github.com/WCRP-CMIP/WCRP-universe.git
# WCRP-CMIP/WCRP-universe
# https://WCRP-CMIP.github.io/WCRP-universe


folders = glob.glob('*/')


for folder in folders:
    print(folder)
    files = glob.glob(folder + '*.json')
    for file in files:
        with open(file, 'r') as f:
            
            js = json.load(f)
            
            if '@type' not in js: 
                continue
            js['id'] = file.split('/')[-1]
            js['type'] = js['@type'].replace('-', '_')
            js['@context'] = f'{pages_url}/src-data/{folder}_context'
            del js['@type'], js['@id']
            
            js = OrderedDict(sorted(js.items()))
            
        
        json.dump(js, open(file,'w') , indent=4)
        # print(file)
        
        
    # context new file
    location = f"{pages_url}/src-data/{folder}"
    context = {
        
        "@context": {
            "@base": location,
            "@vocab": location,
            "id": "@id",
            "type": "@type"
        },
        "@embed": "@always"
        }
                    
    json.dump(context, open(f"{folder}_context.json", 'w'), indent=4)
    
    
    os.system('rm -f ' + folder + '*.jsonld')
       



import glob
import os
import json
from ..locations import compact_url
from ..utils.git import io2url


def main():

    for dir in glob.glob("./src-data/*/"):
        # DO THIS EXTERNALLY

        if not os.path.exists(dir+'_schema_') and not os.path.exists(dir+'_context'):
            # skip all directories without a schema and context
            continue

        contents = sorted([i.split('/')[-1].replace('.json', '')
                          for i in glob.glob(dir+'*.json')])

        # get the existing schema
        schema = json.load(open(dir+'_schema_'))
        schema['id'] = dir.split('/')[-2]

        # Update the files that exist
        # if '$defs' not in schema:
        #     schema['$defs'] = {}

        schema['contains'] = {
            "type": "string",
            "enum": contents
        }

        # Lets update the LD linked references.

        ctx = json.load(open(dir+'_context_'))['@context']

        if not isinstance(ctx, list):
            ctx = [ctx]

        for c in ctx:
            if isinstance(c, dict):
                for key, value in c.items():

                    if isinstance(value, dict) and '@id' in value.values():

                        ref = value["@context"].replace("_context_", "")

                        define = f'A reference object from <a href={ref}> {compact_url(ref)} </a>. '

                        description = define + \
                            value.get('description', "").split('</a>')[-1]

                        schema['properties'][key] = {**schema['properties'].get(key, {}),
                                                     **{"type": schema['properties'][key].get("type", "string"),
                                                         "description": description,
                                                         "$ref": f"{ref}_schema_#/contains"
                                                        }
                                                     }

                # undo earlier mistake
                if key in schema:
                    del schema[key]

                schema['base'] = io2url(c['@base'], path_base='src-data/')

        # from pprint import pprint

        print(schema.keys())

        json.dump(schema, open(dir+'_schema_', 'w'), indent=4)

        print('add schema -> schema.json in production')


if __name__ == "__main__":
    main()




import os
import json
import jsonschema
from jsonschema import validate

from cmipld.utils import read_jsn
from cmipld.repo_info import toplevel


def validate_json(jsn):

    if not isinstance(jsn, dict):
        # if we do not give a file, read this
        jsn = read_jsn(jsn)
    name = os.path.basename(jsn['@id'])

    schema_url = os.path.dirname(jsn['@id']).split(':')[-1]

    schema_loc = f"{toplevel()}/JSONLD/{schema_url}/schema.json"
    # outfile guarantees that we must run this

    schema = read_jsn(schema_loc)

    try:
        validate(instance=jsn, schema=schema)
        print(f"Validation succeeded: {name}")
        return True, f"Validation succeeded: {name}"
    except jsonschema.exceptions.ValidationError as err:
        print("Validation error:", err.message, name)
        return False, "Validation error:\n {err.message}\n RelevantFile: {jsn['@id']}", False





'''
This file starts the server runs all the files in the generate_scripts repository. 
'''

# %%
import cmipld
import importlib
import json
from collections import OrderedDict
import glob
import os
import sys
import re
# from p_tqdm import p_map
import tqdm
from cmipld.utils.offline import LD_server
from cmipld.utils.checksum import version


def write(location, me, data):
    summary = version(data, me, location.split("/")[-1])

    if os.path.exists(location):
        old = cmipld.utils.io.rjson(location)
        if old['Header']['checksum'] == summary['Header']['checksum']:
            return 'no update - file already exists'

    cmipld.utils.io.wjsn(summary, location)


def run(file):
    if file == __file__:
        return

    cmipld.utils.git.update_summary(f'Executing: {file}')

    try:
        # this = importlib.import_module(os.path.abspath(file))

        spec = importlib.util.spec_from_file_location("module.name", file)
        this = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(this)  # Load the module

        processed = this.run(localhost, whoami, repopath, reponame)
        write(*processed)

    except Exception as e:
        cmipld.utils.git.update_summary(f"Error in {file} : {e}")

    return


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Process a file path.")
    parser.add_argument(
        "dir", type=str, help="Path to the generate scripts directory")
    parser.add_argument("repos", type=json.loads,
                        help="JSON string containing repositories")

    args = parser.parse_args()

    print(f"File path provided: {args.dir}")

    # relpath = __file__.replace('__main__.py', '')
    relpath = args.dir

    repo_url = cmipld.utils.git.url()
    io_url = cmipld.utils.git.url2io(repo_url)

    # branch = cmipld.utils.git.getbranch()
    repopath = cmipld.utils.git.toplevel()
    reponame = cmipld.utils.git.getreponame()

    whoami = cmipld.reverse_mapping()[io_url]

    print('-'*50)
    print(f'Parsing repo: {whoami}')
    print(f'Location: {repo_url}')
    print(f'Github IO link: {io_url}')
    print('-'*50)

    ldpath = cmipld.utils.git.ldpath()

    repos = args.repos

    # {
    #     'https://wcrp-cmip.github.io/WCRP-universe/': 'universal',
    #     # 'https://wcrp-cmip.github.io/MIP-variables/': 'variables',
    #     # 'https://wcrp-cmip.github.io/CMIP6Plus_CVs/': 'cmip6plus'
    # }

    localserver = LD_server(repos=repos.items(), copy=[
                            [ldpath, whoami]], override='y')

    localhost = localserver.start_server(8084)
    cmipld.processor.replace_loader(
        localhost, [list(i) for i in repos.items()])

    files = glob.glob(relpath+'*.py')

    for file in tqdm.tqdm(files):
        run(file)

    localserver.stop_server()

    if __name__ == '__main__':
        main()





from pydantic import BaseModel, field_validator
from pydantic import StrictStr, StrictBool, StrictFloat, conlist
from typing import Union, Optional
import json


minlen = 50


class description_field:

    description: StrictStr

    @field_validator('description', mode='after')
    @classmethod
    def description_minlen(cls, value):
        if len(value) <= minlen:
            raise UnicodeError(
                f"Description length == {len(value)}. This is less than the minimum character length of {minlen}. `{value}`")
        return value




import os
import json


def read_jsn(f):
    return json.load(open(f, 'r'))


rjsn = read_jsn


def read_url(url):
    import urllib
    import urllib.request
    try:
        with urllib.request.urlopen(url) as response:
            data = response.read().decode('utf-8')
            json_data = json.loads(data)
            return json_data
    except urllib.error.HTTPError as e:
        err = f"Error: {e.code} - {e.reason}"
        # print(err)
        return None
    except urllib.error.URLError as e:
        err = f"Error: {e.reason}"
        # print(err)
        return None


def wjsn(data, f):
    with open(f, 'w') as file:
        json.dump(data, file, indent=4)


# git reset --hard miptables/jsonld && git clean -fd



from ..locations import mapping
from ..utils import sorted_ctx
from collections import OrderedDict

import glob
import json
import tqdm
import os

from ..locations import reverse_mapping
reverse = reverse_mapping()

repo = os.popen("git remote get-url origin").read().replace('.git',
                                                            '').strip('\n').split('/')[-2:]
base = f'https://{repo[0].lower()}.github.io/{repo[1]}/'
# sort by length of key
mapping = dict(sorted(mapping.items(), key=lambda item: len(item[0])))
mapping['id'] = '@id'
mapping['type'] = '@type'
mapping['entries'] = '@none'

historic = ['id', 'type']


def main():

    # note, an easier, but messier way can achieved by speifying two contexts in the file.
    # @context: [context1, context2]

    ctxs = glob.glob('src-data/*/*_context_')
    print('Updating contexts: to match latest repository prefixes')
    for cx in tqdm.tqdm(ctxs):
        # print(cx)

        item = cx.split('/')[1]

        try:

            data = json.load(open(cx))

            ctx = data['@context']

            # we can ignore one line ctx
            if isinstance(ctx, list):
                if len(ctx) > 1:
                    ctx = ctx[1]
                    # we can ignore one line ctx
                else:
                    ctx = {}

            # this will be removed eventually
            for rm in historic:
                if rm in ctx:
                    # list comprehension to remove historic prefixes from mapping
                    del ctx[rm]

            ctx = OrderedDict(sorted((k, v) for k, v in ctx.items()))

            # ensure the base and vocab are correct
            ctx['@base'] = f"{base}{item}/"
            ctx['@vocab'] = ctx['@base']
            # lets glue together with the global context
            ctx = [f"../_context_", ctx]
            # update the ctx
            data['@context'] = ctx
            data = sorted_ctx(data)

            data['@type'] = [reverse[base]]

            with open(cx, 'w') as f:
                json.dump(data, f, indent=4)

        except Exception as e:
            print(f"Error: {str(e)}")
            print(f"Error with {cx}")
            continue

    # lets write the root repo
    with open('src-data/_context_', 'w') as f:
        data = {"@context": mapping}

        data['@context']['@base'] = base
        data['@context']['@vocab'] = base

        data = sorted_ctx(data)

        json.dump(data, f, indent=4)





from pydantic import BaseModel, field_validator
from pydantic import StrictStr, StrictBool, StrictFloat
from typing import Union, Optional, List
import json

from .stringcheck import hyphenate, maxlen


import re


typestr = re.compile(r'[a-z\-:\d]+')
max_type = 50


class type_field:

    # set field type
    type: Union[str, list[str]]

    type_hyphen = field_validator('type', mode='after')(
        hyphenate)  # Directly apply hyphenate

    type_maxlength = field_validator('type', mode='after')(maxlen(max_type))

    @field_validator('type', mode='after')
    @classmethod
    def type_allowed_char(cls, value):
        for i in value:

            if typestr.fullmatch(i) == None:
                raise UnicodeError(
                    f"Types must only have a-z, lowercase letters and colons (:) to separate the prefixes from the type. Full urls should be provided within the context for clarity. [\"id\": \"{i}\"] has the invalid character: ({set(typestr.sub('',i))}).")
        return value

    @field_validator('type', mode='before')
    @classmethod
    def types_allowed(cls, value):
        ''' 
        # to change the default allowed types use the following in your main code

        def allowed_types(value):
            assert value in ['type1','type2','type3']
            return value

        types_field.types_allowed = allowed_types

        '''
        # this passes by default, we can change this in the main code (see docstring above)
        return value




# python -m cmipld.tests.elements
import cmipld.tests.elements
import cmipld.utils.git as gitutils
from cmipld.utils.git.actions import parse_md, dispatch
import configparser
# from . import activity
import json
import sys
import os
import re


def action():
    issue_body = os.environ.get('ISSUE_BODY')
    print(issue_body)
    conf = parse_md(issue_body)
    print(conf)
    main(conf)


def main(config):
    errors = []

    for section in config.keys():
        try:
            # data
            entry = config[section]
            # library
            section = section.replace('-', "_")
            entrylib = getattr(cmipld.tests.elements, section)
            # class + data
            entryclass = getattr(entrylib, section)()

            # print(entry)

            if not entryclass.create_jsonld(entry, write=True):
                continue

            # otherwise write to file and save repo on git.

            # Set up repo
            gitutils.update_issue_title(f"{entryclass.pullname}")
            branch = f'{entrylib.elementtype}-{entryclass.getid}'
            branch = gitutils.prepare_pull(branch)

            # write to file
            json.dump(entryclass.json, open(entryclass.path, 'w'), indent=4)

            # update issue status
            now = cmipld.utils.get_datetime()
            gitutils.update_issue(
                f'Issue updated: {now} \n\nBranch: {branch} \n\n ```json \n {entryclass.jsonstr} \n```', False)

            # commit the file.
            print(f"Committing {entryclass.path}")
            author = os.environ.get('ISSUE_SUBMITTER', 'cmip-ipo')
            gitutils.commit_one(
                entryclass.path, author, f"New entry {entryclass.getid} to the {entrylib.elementpath} LD file", branch)

            gitutils.pull_req(
                branch, author, entryclass.jsonstr, entryclass.pullname)

        except ModuleNotFoundError as e:
            errors.append(
                f"No such module {section} in cmipld. Please check the congifuration file template.")
            continue

        if errors:
            print(errors)

    # git add one file.
    # git commit with user
    # git push


# conf = '''
# [institution]
#     Acronym = CMIP-IPO
#     Full_Name = CMIP Model Intercomparison Project Office
#     ROR = 000fg4e24

#     # only change the item below to "update" if you are submitting a correction.
#     action = new

# [activity]
#     Name = CMIP
#     Long_Name =  Coupled Model Intercomparison Project
#     URL = https://wcrp-cmip.org

#     # only change the item below to "update" if you are submitting a correction.
#     action = new

# [sub-experiment-id]
#     name =  new-sub-experiment
#     description = A sample sub-experiment id

#     # only change the item below to "update" if you are submitting a correction.
#     action = new
# '''




import json
from pyld import jsonld
import jmespath
import re
import os
import pkg_resources


def get_frame(base, file):
    '''Get frame from saved collections'''
    # sanitise path
    path = os.path.normpath(f'{base}/{file}.json')
    # with open(__file__.replace('__init__.py','examples/')+path,'r') as f:
    # print(__name__)
    # this works from the current directory
    with pkg_resources.resource_stream(__name__, f'examples/{path}') as f:
        return json.load(f)


def direct(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AttributeError as e:
            if "object has no attribute 'json_string'" in e.__str__():
                print(
                    '!! Error handling has corrected the error.\n In the future, please ensure that parsing functions are called using "clean(process=[...])" method or between a start and end method.\n For more information consult the documentation.')
                args[0].clean(process=[func.__name__])
                return args[0]

            # print(f"An error occurred in {func.__name__}: {e}")
            # # Handle error, e.g., return a placeholder string
            # return "Error occurred"
    return wrapper


class Frame:
    def __init__(self, source, frame, nograph=True):

        # if isinstance(source, list):
        #     source = {"@context":{},"@graph":source}
        # elif isinstance(source, str):
        #     source = json.loads(source)

        self.source = source
        self.frame = frame

        # "@embed": "@always"

        if "@embed" not in frame:  # make sure each instance is embedded.
            self.frame["@embed"] = "@always"

        if nograph:  # usually used if framing an id directly
            self.data = self.graph_only(jsonld.frame(source, frame))
        else:
            self.data = jsonld.frame(source, frame)

    def __str__(self):
        return json.dumps(self.data, indent=2)

    # we do not need to call this as a function
    @property
    def print(self):
        from pprint import pprint
        pprint(self.data)

    @property
    def json(self):
        return self.data

    # @property

    def clean(self, process=None):

        if not process:
            process = ['rmld', 'rmnull', 'untag', 'flatten']

        self.start

        for function in process:
            self.__getattribute__(function)

        self.end
        return self

    @property
    def filter_empty(self):

        # self.data = {k:v for k,v in self.data.items() if v}
        self.data = [v for v in self.data if v]

        return self

    @property
    def clean_cv(self):
        return self.clean(['rmld', 'missing', 'untag', 'lower', 'flatten'])

    # @staticmethod
    @property
    def start(self):
        self.json_string = json.dumps(self.data, indent=2)
        return self

    # @staticmethod
    @property
    def end(self):
        self.data = json.loads(self.json_string)
        del self.json_string
        return self

    @property
    @direct
    def untag(self):
        self.json_string = re.sub(
            r'"(?!mip:|@)[^@":]*:([^@"]*?)":', r'"\1":', self.json_string)
        return self

    @property
    @direct
    def rmld(self):
        self.json_string = re.sub(
            r'("@[^"]*":\s*".*?"(?:,)?\s*)', '', self.json_string)
        return self

    @property
    @direct
    def rmnull(self):
        self.json_string = re.sub(
            r'[\s]*"(.*?)":\s*null*[,\s]*|[,\s]*"(.*?)":\s*null*[\s]*', '', self.json_string)
        return self

    @property
    @direct
    def rmnoparent(self):
        # Remove key-value pairs where the value is 'no parent'
        self.json_string = re.sub(
            r'[\s]*"(.*?)":\s*"no parent"[,\s]*',  # Pattern to match
            '',  # Replacement (remove the matched part)
            self.json_string
        )
        # Clean up any trailing commas or whitespace
        self.json_string = re.sub(r',\s*(?=[}\]])', '', self.json_string)
        return self

    @property
    @direct
    def flatten(self):
        self.json_string = re.sub(
            r'{\s*"([^"]*?)":\s*"(.+)"\s*}', r'"\2"', self.json_string)
        return self

    @property
    @direct
    def lower(self):
        self.json_string = re.sub(
            r'(?<=")([^"]*?)-(.*?)(?=":)', lambda m: m.group(0).replace('-', '_'), self.json_string)
        return self

    @property
    @direct
    def missing(self):
        self.json_string = re.sub(
            r'\{\s*"@id"\s*:\s*"([^"]+)"\s*\}', r'"Missing Link: \1"', self.json_string)
        return self

    @staticmethod
    def graph_only(json_object):
        return json_object.get('@graph', {})

    # @staticmethod
    # def regex_chain(json_string):
    #     return RegexChain(json_string)

    @staticmethod
    def print_state(json_object):
        print(json_object)
        return json_object

    @staticmethod
    def search(json_object, jmespath_query):
        return jmespath.search(jmespath_query, json_object)

    @staticmethod
    async def validate_jsonld(json_object):
        try:
            expanded = jsonld.expand(json_object)
            return True
        except jsonld.JsonLdError as e:
            print(f"JSON-LD validation error: {e}")
            return False

    @staticmethod
    async def compact_jsonld(json_object, context):
        try:
            compacted = jsonld.compact(json_object, context)
            return compacted
        except jsonld.JsonLdError as e:
            print(f"JSON-LD compaction error: {e}")
            return None

    # @property
    # def rmempty(self):
    #     print(self.data)
    #     self.data = dict([i for i in self.data if i.values()])

    #     return self

    def filterkeys(self, keys, overwrite=False, missing=None):
        # assert not set(keys) - set(self.data)
        output = []
        for d in self.data:
            dummy = {}
            for k in keys:
                dummy[k] = d.get(k, missing)
            output.append(dummy)

        if overwrite:
            self.data = output
            return self
        return output

    def key_value(self, key='name', value=None):
        return key_value(self.data, key, value)

    def key_only(self, key='name', default='missing'):
        return key_only(self.data, key, default)

    def value_only(self, key):
        return value_only(self.data, key)

    # def key_str(self, key,value):
    #     return key_str(self.data,key,value)


####################
# Utility functions
####################

def key_value(data, key='name', value=None):
    return {
        (k := d.pop(key, f'keyNotFound[{key}]')): (d.get(value, f'valueNotFound[{value}]') if value else d)
        for d in data
    }


def value_only(data, key):
    try:
        return sorted(list(set([d.get(key) for d in data if d.get(key)])))
    except TypeError as e:
        return list([d.get(key) for d in data if d.get(key)])


# def key_str(data,key,value):
#     return dict([[d.get(key),d.get(value)] for d in data])

def key_only(data, key='name', default='missing'):
    if isinstance(data[0], str):
        # if single value do not fetch key
        return data
    return sorted(list(set([x.get(key, default) for x in data if x])))




from pydantic import BaseModel, field_validator
from pydantic import StrictStr, StrictBool, StrictFloat
from typing import Union, Optional
import json

from .ror import ror_field


from ..components import id_field, type_field


class institution(BaseModel, id_field, type_field, ror_field):

    @field_validator('type', mode='after')
    @classmethod
    def type_contains(cls, value):
        assert 'wcrp:organisation' in value
        assert 'wcrp:institution' in value
        assert 'wcrp:consortium' not in value
        return value

    def json(self):
        return self.model_json_dump()

# Example usage
# a = EMD(**{'num':'1','id':'ts-t'})




from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic import StrictStr, StrictBool, StrictFloat
from typing import Union, Optional, List
import json
import os
from cmipld import mapping, jsonld


from ..components import id_field, type_field, validate_date,  description_field


class experiment_model(BaseModel, id_field, type_field, description_field):

    tier: int
    minimum_number_of_years: int = Field(..., alias="minimum-number-of-years")
    start_date: Optional[StrictStr] = Field(..., alias="start-date")
    # branch_date: Optional[StrictStr] = Field(..., alias="branch-date")
    model_realms: Optional[list[dict]] = Field(..., alias="model-realms")
    description: StrictStr
    parent_experiment: List[StrictStr] = Field(..., alias="parent-experiment")

    @field_validator('type', mode='after')
    @classmethod
    def type_contains(cls, value):
        assert 'wcrp:experiment' in value, 'type should contain wcrp:experiment'
        return value

    @field_validator('tier', mode='after')
    @classmethod
    def tier_val(cls, value):
        assert value in [1, 2, 3], 'tier should be 1,2, or 3'
        return value

    # @field_validator('start_date', mode='after')
    # @classmethod
    # def s_date(cls, value):
    #     if value != 'none':
    #         validate_date(value)
    #     return value


    @model_validator(mode='after')
    @classmethod
    def s_date(cls, values):

        if values['start-date'] != 'none':
            validate_date(values['start-date'])
        else: 
            if values["minimum-number-of-years"] > 0:
                raise ValueError('If no start date is provided, a minimum number of years entry must be supplied.')
            
        return values


    # @field_validator('branch_date', mode='after')
    # @classmethod
    # def s_date(cls,value):
    #     if value != 'none':
    #         validate_date(value)
    #     return value

    @field_validator('model_realms', mode='after')
    @classmethod
    def ld_realms(cls, value):

        urlbase = mapping['universal']
        for i in value:

            url = f'{urlbase}source-type/{i['id']}'
            try:
                jsonld.expand(url)
            except:
                raise ValueError(
                    f"Invalid model realm: {i['id']}. Please check {url} exists and the data is correct. If missing, this may need to be registered. ")
            # jsonld.compact(url,url)['label']
        return value

    @field_validator('parent_experiment', mode='after')
    @classmethod
    def parent_e(cls, value):
        assert len(value) == 1, 'parent experiment should be a single value'
        if value[0] == 'none':
            return value

        dir = 'src-data/experiment/'
        if value[0].lower() not in [f.replace('.json', '') for f in os.listdir(dir)]:

            raise ValueError(f"Parent experiment {value} not found in {dir}")

        print('can a parent experiment be exist in another project?')

        return value

    def json(self):
        return self.model_json_dump()

# Example usage
# a = EMD(**{'num':'1','id':'ts-t'})
